#############################################################################
#
# Package: The Canon Tensor Package
#
# Language: Maple 10.
#
# Description: A fast manipulator of generic scalar polynomial expressions 
#   formed from the Riemann tensor of a four-dimensional metric-compatible 
#   connection. Manipulation of generic tensor expressions. Tensor may have
#   permutation symmetries.
#
# Developers: Renato Portugal
#             Laboratorio Nacional de Computacao Cientifica (LNCC)
#                            and
#             Jose Maria Martin-Garcia
#             Consejo Superior de Investigaciones Cientificas (CSIC)
#
# Web page: http://www.lnc.br/~portugal/Canon.html
#
# Contact and bug report: Portugal@Lncc.Br
#
#############################################################################
#
# COPYRIGHT NOTICE
#
# The Canon tensor package has all rights reserved to the developers. The 
# package and documentation are provided as it is, without guarantee of support 
# or maintenance. We are not liable for any damages resulting in any way from 
# their use.
# 
# Everyone is granted permission to copy, modify and redistribute the Canon
# package and its documentation, provided that, (1) all copies contain this
# notice in the main program file and in the documentation, (2) no charge 
# is made for this software or works derived from it.
#
#############################################################################
#
# Structure of this file
#
# 1. INITIALIZATION OF CANON TENSOR PACKAGE (Supersedes OLD CANON)
# 
# 2. USER PROCEDURES
#
# 3. INTERNAL COMPUTATIONAL-GROUP-THEORY PROCEDURES
#
# 4. TENSOR PROCEDURES
#
# 5. TYPES (tensor)
#
# 6. INTERFACE
#
# 7. GENERAL TOOLS
#


unprotect( 'Invar' ):

Invar := module()

description "The Invar tensor package on top of the Canon tensor package";

export 	RiemannSimplify,
	AbsorbMetric, 
	PermToInv,
	MaxIndex,
	MaxDualIndex,
	RiemannToPerm,
	PermToRiemann,
	InvToPerm,
	WeylToRiemann,
	RiemannToWeyl,
	RicciToRiemann,
	RiemannToRicci,
	RiemannToInv,
	InvToRiemann,
	TensorDefine,
	Canonical,
	TensorPrint,
	TensorUnPrint
	;

local CanArg,EnvCanonFreeInds,
 	setup, cleanup, 
	`Canon/RiemannSimplify`, `Canon/AbsorbMetric`, 
	`Canon/AbsorbMetric/aux`, `Canon/AbsorbMetric/aux/rec`, 
	`Canon/PermToInv`, `Canon/PermToInv/aux`, `Canon/MaxIndex`, 
	`Canon/RiemannToPerm`, `Canon/PermToRiemann`, `Canon/WeylToRiemann`, 
	`Canon/RiemannToWeyl`, `Canon/RicciToRiemann`, 
	`Canon/RiemannToRicci`, `Canon/RiemannToInv`, `Canon/InvToRiemann`, 
	`Canon/InvToRiemann/1`, `Canon/MkNewDumInds`, `Canon/CanonicalOne`, 
	`Canon/position`, `Canon/power`, `Canon/prod`, `Canon/prod/nosign`, 
	`Canon/g_power`, `Canon/inv`, `Canon/invcyc`, `Canon/min`, 
	`Canon/SortPermutations`, `Canon/SortPermutations/aux`, 
	`Canon/orbit`, `Canon/orbit1`, `Canon/orbit2`, `Canon/allorbits`, 
	`Canon/allorbits0`, `Canon/trace`, `Canon/singcoset`, 
	`Canon/doubcoset`, `Canon/doubcoset/F1`, `Canon/doubcoset/sgd`, 
	`Canon/SGS_dummy`, `Canon/Canonical/Image`, `Canon/Canonical/g`, 
	`Canon/isingroup`, `Canon/schreier-sims`, `Canon/schreier-sims/aux`, 
	`Canon/TensorDefine`, Canonical1, Canonical2, `Canon/Canonical/aux`, 
	`Canon/Canonical/SplitBlocks`, `Canon/Canonical/CanDummy`, 
	`Canon/Canonical/TensorProd`, `Canon/Canonical/Tensor`, 
	`Canon/Canonical/CanonTensorToProd`, 
	`Canon/Canonical/CanonTensorToProd/1`, 
	`Canon/has_sym`, `Canon/Canonical/MkSym`, `Canon/Canonical/sym_displ`, 
	`Canon/Canonical/NewDummies`, `Canon/indices`, `Canon/indices/rec`, 
	`Canon/AllDummies`, `Canon/FreeIndices`, `Canon/FreeIndices/aux`, 
	`Canon/Canonical/expand`, `Canon/Canonical/Perm`, 
	`Canon/Canonical/MkSymRieProd`, `Canon/Canonical/MkSymRieProd*epsilon`, 
	`Canon/TensorPrint`, `_TENSOR/F`, `Canon/TensorUnPrint`, 
	`Canon/Canonical/Remove_SCALAR`, `Canon/positions`, 
	`Canon/selectrepeated`, `Canon/tools/sort`, `Canon/abs`, 
	`Canon/tools/joinsets`
	; 

option 	package,
	load = setup,
	unload = cleanup;

global 	TensorTable,CanonSimp,SpaceDimension,DummyIndices;



#############################################################################
#############################################################################
#
# 1. INITIALIZATION OF CANON

CanArg := ['CanonSimp','SpaceDimension','DummyIndices',
           '_EnvCanonRiemannToInv','_EnvCanonRiemannToPerm'];

setup := proc()

global `Invar/version`,`Canon/version`,
	CanonSimp,TensorTable,SpaceDimension,
	DummyIndices, 
	_TENSOR, `print/_SCALAR`, `print/_FREE`, `print/_TENSOR`,
	`type/cycle`, `type/permutation`, `type/sigperm`, `type/genperm`, 
	`type/tensorsymmetry`,`type/constanttensor`, `type/scalarinvariant`, 
	`type/CanonTensor`, `type/contravarianttensor`, `type/covarianttensor`, 
	`type/nameindex`, `type/contravariantnameindex`, 
	`type/covariantnameindex`, `type/contravariantindex`, 
	`type/covariantindex`, `type/numericindex`, 
	`type/contravariantnumericindex`, `type/covariantnumericindex`;


`Invar/version` := proc()
options `Copyright 2003-2006 by R Portugal. All rights reserved`;
"Invar tensor package version 1.0, Nov, 23, 2006, Maple 9,10"
end proc:

`Canon/version` := proc()
options `Copyright 2003-2006 by R Portugal. All rights reserved`;
"Canon tensor package version 2.0, Sep, 06, 2006, Maple 7,8,9,10"
end proc:

# Global Variables 
# Convention: The name of all global variables starts with "Canon"
# Globals have the drawback with respect to using option remember.
#
# List of global variables:
# USER LEVEL:
# CanonSimp,TensorTable,SpaceDimension,DummyIndices
# INTERNAL LEVEL:
# CanonTensor,CanonCanDummyInds,CanonExtraDummyInds,
#
# Tables of invariants:
# Invars,DualInvars,CyclicEq,DualCyclicEq,DimDepEq,DualDimDepEq,DualDualEq
# 
# These global variable controls the construction of the inverse tables:
# `Canon/Invars`,`Canon/DualInvars`,`Canon/Invars/flag`,`Canon/DualInvars/flag`

_EnvCanonRiemannToInv := false;
_EnvCanonRiemannToPerm := false;
DummyIndices := {};

#
# CanonSimp can be assigned to a simplifying function. It is applied 
# inside command Canonical when the tensor expression have canonical 
# names for the dummy indices.
# Example: CanonSimp := factor;
CanonSimp := x -> x;

#
#
# CanonMetricSymmetry - symmetry of the metric
#
# 1 - symmetric   -1 - antisymmetric    0 - no symmetries
#
#CanonMetricSymmetry := 1:  # Allowable values = {1,0,-1} 

#
# Dimension of the manifold for trace of symmetric metric
#
SpaceDimension := 4:

#
# TensorTable - table of the symmetries of the tensors defined
#                      by TensorDefine.
#
TensorTable := table():

#
# CanonTensor is used as the name of internal tables. No 
# assignment is performed.
#
protect(CanonTensor):

#
# _FREE is a inert function used in internal procedures. It does not
# appear in outputs.
#
protect('_FREE'):
protect('_SCALAR');

#
# The symbols `` and ` ` are used in the output (interface)
#
protect(``):
protect(` `):

#
# Protect table names
#
#protect('Invars','DualInvars','CyclicEq','DualCyclicEq',
#        'DimDepEq','DualDimDepEq',
#        'DualDualEq'):

# Macros 



#############################################
#
# Built-in Tensors 
#
# This block must come after the command `Canon/TensorDefine`

protect('R','epsilon','C','g'):
#
# 1 - The Riemann tensor
#
`Canon/TensorDefine`(
    'R', 4,
       {[-1,[[1,2]]],[-1,[[3,4]]],[[1,3],[2,4]]},
         [$1..4]):
#
# 2 - The Ricci tensor
#
`Canon/TensorDefine`(
    'R', 2,{[[1,2]]},[$1..2]):
#
# 3 - The Ricci Scalar
#
`Canon/TensorDefine`(
    'R', 0, {}, []):

#
# 4 - The totally antisymmetric tensor
#
`Canon/TensorDefine`(
    'epsilon', 4,
       {[-1,[[1,2]]],[-1,[[2,3]]],[-1,[[3,4]]]},
         [$1..4]):
#
# 5 - The Metric tensor
#
`Canon/TensorDefine`(
    g, 2,{[[1,2]]},[$1..2]):

###Metric := 'g':
#
# 1 - The Weyl tensor
#
`Canon/TensorDefine`(
    'C', 4,
       {[-1,[[1,2]]],[-1,[[3,4]]],[[1,3],[2,4]]},
         [$1..4]):

#
# End of Built-in Tensors 
#
##################################################

# 5. TYPES (tensor)


`type/constanttensor` := proc(x)
options remember,system,`Copyright 1995-2003 by R. Portugal. All right reserved.`;
type(x,'indexed') and type(op(0,x),'symbol') and
type([op(x)],list('{nameindex,posint}'))
end proc:

`type/scalarinvariant` := proc(x)
options remember,system,`Copyright 1995-2003 by R. Portugal. All right reserved.`;
type(x,indexed) and map(op,`Canon/indices`(x)[2..3])=[]
end proc:

`type/CanonTensor` := proc(x)
options remember,system,`Copyright 2003 by LRU Manssur and R. Portugal. All right reserved.`;
type(x,indexed) and type(op(0,x),indexed) and op([0,0],x)='CanonTensor'
end proc:

`type/contravarianttensor` := proc(x)
options `Copyright 1997 by R. Portugal. All right reserved.`;
type(x,'constanttensor') and type([op(x)],list('contravariantindex'))
end proc:

`type/covarianttensor` := proc(x) 
options `Copyright 1997 by R. Portugal. All right reserved.`;
type(x,'constanttensor') and type([op(x)],list('covariantindex'))
end proc:

`type/nameindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,'algebraic') and not type(x,'numericindex') and 
(type(x,'symbol') or type(-x,'symbol')) 
end proc:

`type/contravariantnameindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,'symbol') 
end proc:

`type/covariantnameindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,algebraic) and type(-x,'symbol')  
end proc:

`type/contravariantindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,'{symbol,contravariantnumericindex}') 
end proc:

`type/covariantindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,algebraic) and type(-x,'{symbol,contravariantnumericindex}')  
end proc:

`type/numericindex` := proc(x)
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if not type(x,'algebraic') then 'false'
elif not type(x,integer) then 'false'
elif x=0
  then 'false'
else true 
end if
end proc:

`type/contravariantnumericindex` := proc(x)
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if not type(x,'algebraic') then 'false'
elif not type(x,nonnegint) then 'false'
elif x=0
  then 'false'
else true 
end if
end proc:

`type/covariantnumericindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if not type(x,'algebraic') then 'false'
elif not type(-x,nonnegint) then 'false'
elif x=0
  then 'false'
else 'true' 
end if
end proc:


##############
# Type cycle #
##############

`type/cycle` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x,list(posint)) and nops(x)>1

#
# tests for repetitions
#
and nops(x)=nops({op(x)})

#
# tests for correct order
#
and min(op(x))=x[1]

end proc:



####################
# Type permutation #
####################

`type/permutation` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x, list('cycle'))

#
# tests for repetitions among cycles
#

and convert(map(nops,x),`+`)=nops({op(map(op,x))})

#
# tests for correct order
#

and sort(map2(op,1,x))=map2(op,1,x)

end proc:



###########################
# Type signed permutation #
###########################

`type/sigperm` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x,[1,'permutation']) or
type(x,[-1,'permutation'])

end proc:



############################
# Type general permutation #
############################

`type/genperm` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x,'permutation') or type(x,'sigperm')

end proc:



######################
# Type tensorsymmetry #
######################

`type/tensorsymmetry` := proc(x) #-> set of permutations/signedperm

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x, set('genperm'))

end proc:


# 6. INTERFACE

_TENSOR := proc()
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
if type(args[1],`*`) then 
       map('_TENSOR',args[1],args[2..-1])
  else '_TENSOR'(args)
end if
end proc:

protect('_TENSOR');

`print/_TENSOR` := proc(x)
options remember,system,`Copyright 1995 by R. Portugal. All right reserved.`;
if type(x,'indexed') then
     if nops(x)=0 then 
            op(0,x)
     else convert([readlib('`tools/gensym`')(op(0,x)),
                                   op(map(`_TENSOR/F` ,[op(x)]))],`*`)
     end if
elif type(x,'function') then 
     if type(op(0,x),'indexed') then 
              convert([readlib('`tools/gensym`')(op(0,op(0,x))),
                 op(map(`_TENSOR/F` ,[op(op(0,x))])),
                    readlib('`tools/gensym`')(``)(
                       op(map(procname,[op(x)])))],`*`)
     else x
     end if
else x
end if
end proc:

`print/_SCALAR` := proc(x) ``(x) end:

`print/_FREE` := proc(x) ``(x) end:

NULL;
end:

######## end setup ##########


cleanup := proc()
global `Invar/version`,`Canon/version`,
	CanonSimp,TensorTable,SpaceDimension,
	DummyIndices, 
	_TENSOR, `print/_SCALAR`, `print/_FREE`, `print/_TENSOR`,
	`type/cycle`, `type/permutation`, `type/sigperm`, `type/genperm`, 
	`type/tensorsymmetry`,`type/constanttensor`, `type/scalarinvariant`, 
	`type/CanonTensor`, `type/contravarianttensor`, `type/covarianttensor`, 
	`type/nameindex`, `type/contravariantnameindex`, 
	`type/covariantnameindex`, `type/contravariantindex`, 
	`type/covariantindex`, `type/numericindex`, 
	`type/contravariantnumericindex`, `type/covariantnumericindex`;

unprotect(CanonTensor,'_FREE','_SCALAR','_TENSOR',``,` `,
	'Invars','DualInvars','CyclicEq','DualCyclicEq',
        'DimDepEq','DualDimDepEq','DualDualEq'):

unassign('
	`Invar/version`,`Canon/version`,
	CanonSimp,TensorTable,SpaceDimension,
	DummyIndices, 
	Invars,DualInvars,CyclicEq,DualCyclicEq,
        DimDepEq,DualDimDepEq,DualDualEq,
	CanonCanDummyInds,CanonExtraDummyInds,
	`Canon/Invars`,`Canon/DualInvars`,
	`Canon/Invars/flag`,`Canon/DualInvars/flag`,
	_TENSOR, `print/_SCALAR`, `print/_FREE`, `print/_TENSOR`,
	`type/cycle`, `type/permutation`, `type/sigperm`, `type/genperm`, 
	`type/tensorsymmetry`,`type/constanttensor`, `type/scalarinvariant`, 
	`type/CanonTensor`, `type/contravarianttensor`, `type/covarianttensor`, 
	`type/nameindex`, `type/contravariantnameindex`, 
	`type/covariantnameindex`, `type/contravariantindex`, 
	`type/covariantindex`, `type/numericindex`, 
	`type/contravariantnumericindex`, `type/covariantnumericindex`
	');

NULL;
end proc:

#############################################################################



#############################################################################
#############################################################################
#
# 2. USER PROCEDURES
#
# Convention: 
#    (1) checks arguments extensively
#    (2) removes the interface
#    (3) calls the respective `Canon/command`
#    (4) puts the interface back
#


#
# RiemannSimplify(expr, opt)
#
#    Symplifies a tensor expression
#
# Tensor is an indexed variable: 
#     e.g. R[a,b,-c,-d]
# where R is the tensor name
#       a,b are contravariant indices
#       -c,-d are covariant indices
# 
# Input: expr - tensor expression
#        option - 
#
# Output: 
#   An expression
#

RiemannSimplify := proc(expr, opt::{identical(DDI),identical(Cyclic)})

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

if nargs=1 then
	return RiemannSimplify(expr,'DDI')
end if;

`Canon/TensorPrint`(
	`Canon/Canonical/Remove_SCALAR`(
		`Canon/RiemannSimplify`( `Canon/TensorUnPrint`(expr), opt ) ) )

end proc:



`Canon/RiemannSimplify` := proc(expr, opt)
local S,expr1,i,sr,S_expr,T;
options `Copyright 1995-2006 by R Portugal. All rights reserved`;
#
# Absorb Metric
#
expr1 := `Canon/AbsorbMetric`(expr);

#
# Simplify contractions of the Weyl tensor
#
S := select(x->nops(map(`Canon/abs`,{op(x)}))<4,
	indets(expr1,'C[anything,anything,anything,anything]') );
expr1 := subs(map(x->x=0,S),{'_FREE(0)'=0,'_SCALAR(0)'=0},expr1);

#
# Ricci -> Riemann, and canonicalizes, and  Riemann -> RInv and DualRInv
#
expr1 := `Canon/RiemannToInv`(expr1);

#
# cyclic eqs
#
if member(opt,'{DDI,Cyclic}') then
	expr1 := subs(
		map(op,{seq(entries(CyclicEq[i]),i=2..7)}),
		map(op,{seq(entries(DualCyclicEq[i]),i=2..5)}),
			expr1);
end if;

if member(opt,'{DDI}') then
	#
	# dimension dependent eqs
	#
	expr1 := subs(
		map(op,{seq(entries(DimDepEq[i]),i=3..7)}),
		map(op,{seq(entries(DualDimDepEq[i]),i=3..5)}),
			expr1);

	#
	# reduction of non duals in terms of duals eqs
	#
	expr1 := subs(
		map(op,{seq(entries(DualDualEq[i]),i=4..7)}),expr1);
end if;

#
# Riemann -> Ricci
#
expr1 := `Canon/RiemannToRicci`(expr1);

#
# Cyclic symmetries for Riemann tensor with free indices
#
if member(opt,'{DDI,Cyclic}') then

	S := indets(expr1,R[anything,anything,anything,anything]);

	if S<>{} then 
		S_expr := Canonical2(expr1);

		sr := NULL;
		for i in S do
			S := {MonomialsWithTensor(expr1,i)};
			sr := sr,subs(i=Cyclicize(i),S);
		end do;
		sr := map(op,{sr});

		for i to nops(sr) do T[i] := Canonical2(sr[i]) end do;

		expr1 := simplify(S_expr[2],{seq(T[i][2]=0,i=1..nops(sr))});

	        expr1 := subs(S_expr[1],seq(T[i][1],i=1..nops(sr)),expr1);

		expr1 := subs( map(
			x->x=`Canon/Canonical/CanonTensorToProd`(x),
				indets(expr1,'CanonTensor')),
					expr1);
	end if;

end if;

Canonical1(expr1,map(eval,CanArg));

end proc:



#
# raises or lowers indices contrated by the metric
#
AbsorbMetric := proc(expr, opt)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

`Canon/TensorPrint`(
	`Canon/Canonical/Remove_SCALAR`(
		`Canon/AbsorbMetric`( `Canon/TensorUnPrint`(expr) ) ) )

end proc:


`Canon/AbsorbMetric` := proc(expr)
local expr1,S;

if not hastype(expr,'g[anything,anything]') then
	return expr
end if;

expr1 := Canonical1(expr,map(eval,CanArg));

S := select(has, indets(expr1, 
	'{_FREE(anything),_SCALAR(anything)}'), 'g');  # Metric

subs( map(x -> x=`Canon/AbsorbMetric/aux`(x), S), expr1 );

end proc:


`Canon/AbsorbMetric/aux` := proc(expr)
local z,e1,e2,res;

e1,e2 := selectremove(type,op(expr)*z,g[nameindex,nameindex]);  # Metric

e2 := subs(z=1,e2);

if e2=1 then
	if type(expr,'_SCALAR(anything)') then
		SpaceDimension
	else
		g[op(convert(map(op,[op(e1)]),`+`))]   # Metric
	end if

else
	if type(expr,'_SCALAR(anything)') then
		'_SCALAR'(`Canon/AbsorbMetric/aux/rec`(e1,e2))
	else
		res := `Canon/AbsorbMetric/aux/rec`(e1,e2);
		if `Canon/AllDummies`(res)={} then
			res
		else
			'_FREE'(res)
		end if;
	end if;

end if;

end proc:



`Canon/AbsorbMetric/aux/rec` := proc(e1,e2)
local g1,i,j,e1_;

if e1=1 then
	e2

else
	if type(e1,indexed) then
		g1 := e1;
		e1_ := 1;
	else
		g1 := op(1,e1);
		e1_ := subsop(1=1,e1);
	end if;
	i := op(1,g1); j := op(2,g1);
	if has([e1_,e2],`Canon/abs`(i)) then
		procname(op(subs(
			`if`(type(i,symbol),i=-j,-i=j),
					 [e1_,e2] ) ))
	else
		procname(op(subs(
			`if`(type(j,symbol),j=-i,-j=i),
					 [e1_,e2] ) ))
	end if;

end if;

end proc:


#
# PermToInv
#
# g is a canonical permutation
#
# dual: false - non dual invariants
#       true -  dual invariants
#
#  ([[2, 3], [4, 7, 6]],false) -> RInv[2,2] 
#
#  ([[2,3,5],[4,7,9,6],[8,11,10]],true) -> DualRInv[2,3]
#
PermToInv := proc(g::genperm,dual::boolean)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

if nargs<>2 then
	error "expecting 2 arguments: g, dual, got ",nargs
end if;

`Canon/PermToInv`(g,dual)

end proc:


`Canon/PermToInv` := proc(g,dual)
local degree,i;

degree := ceil(max(op(indets(g,integer)))/4);

if degree>4 then
	return `Canon/PermToInv/aux`(g,dual,degree)
end if;

if dual then
    degree := degree-1;
    for i to `Canon/MaxIndex`(DualInvars[degree]) do
	if DualInvars[degree][i]=g then
		return DualRInv[degree,i]
	end if;
    end do;
else
    for i to `Canon/MaxIndex`(Invars[degree]) do
	if Invars[degree][i]=g then
		return RInv[degree,i]
	end if;
    end do;
end if;
return g
end proc:


`Canon/PermToInv/aux` := proc(g,dual,deg)
local res,i,degree;
global  `Canon/Invars`,`Canon/DualInvars`,
	`Canon/Invars/flag`,`Canon/DualInvars/flag`;

if dual then

    degree := deg-1;

    if `Canon/DualInvars/flag`[degree] = true then

	res := `Canon/DualInvars`[op(g)];

	if type(res,DualRInv[degree,anything]) then
		return res
	else
		return g
	end if;

    end if;

    for i to `Canon/MaxIndex`(DualInvars[degree]) do
	`Canon/DualInvars`[op(DualInvars[degree][i])] := DualRInv[degree,i];
    end do;

    `Canon/DualInvars/flag`[degree] := true;

else

    degree := deg;

    if `Canon/Invars/flag`[degree] = true then

	res := `Canon/Invars`[op(g)];

	if type(res,RInv[degree,anything]) then
		return res
	else
		return g
	end if;

    end if;

    for i to `Canon/MaxIndex`(Invars[degree]) do
	`Canon/Invars`[op(Invars[degree][i])] := RInv[degree,i];
    end do;

    `Canon/Invars/flag`[degree] := true;

end if;

`Canon/PermToInv/aux`(g,dual,deg)

end proc:


#
# MaxIndex and MaxDualIndex
#
# degree - Number of Riemann tensors
#
MaxIndex := proc(degree::posint)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

if nargs<>1 then
	error "expecting 1 argument: degree got ", args
end if;
`Canon/MaxIndex`(Invars[degree])
end proc:


MaxDualIndex := proc(degree::posint)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

if nargs<>1 then
	error "expecting 1 argument: degree got ", args
end if;
`Canon/MaxIndex`(DualInvars[degree])
end proc:



`Canon/MaxIndex` := proc(TableName)

if type(TableName,table) then
	max(op(map(op,{indices(TableName)})))

else
	0

end if;

end proc:



#
# RiemannToPerm
#
# There is no RiemannToPerm yet. One has to convert the invariant to
# RInv or DualRInv notation an then employ InvToPerm
#
RiemannToPerm := proc(expr)
local i,expr1,S,e1,e2;
options `Copyright 1995-2006 by R Portugal. All rights reserved`;

expr1 := `Canon/TensorUnPrint`(expr);

if not hastype(expr,'specindex(anything,R)') then
	error "expecting an expression with Riemann tensors, got ",expr
elif nops(`Canon/indices`(expr1)[2])>0 then
	error "expecting a scalar, got ",expr
end if;

expr1 := Canonical1(`Canon/RicciToRiemann`(expr1),map(eval,CanArg));

if expr1=0 then return 0 end if;

for i in R[]*expr1 do
  if hastype(i,specindex(nameindex,epsilon)) then
    S := indets(i,specindex(nameindex,epsilon));
    if nops(S)>1 then 
       error "expecting an expression with one epsilon tensor, got ",i
    end if;
    i := subs(op(S)=1,i);
  end if;
  if not type(i,'{-1,R[],R[]^posint,
    R[nameindex,nameindex,nameindex,nameindex],
    `*`(R[nameindex,nameindex,nameindex,nameindex]),
    _SCALAR(R[nameindex,nameindex,nameindex,nameindex]),
    _SCALAR(`*`(R[nameindex,nameindex,nameindex,nameindex])),
    _SCALAR(R[nameindex,nameindex,nameindex,nameindex])^posint,
    _SCALAR(`*`(R[nameindex,nameindex,nameindex,nameindex]))^posint
			}') then
    error "expecting an expression that can be converted to a permutation, got ",
           i
  end if;
end do;

if not has(expr1,'_SCALAR') and type(expr1,`*`) then
	e1,e2 := selectremove(type,expr1,indexed);
	if has(e1,R[]) then
		expr1 := '_SCALAR'(e1/R[])*R[]*e2
	else
		expr1 := '_SCALAR'(e1)*e2
	end if;
end if;

`Canon/TensorPrint`( `Canon/RiemannToPerm`( expr1 ) )

end proc:


#
# `Canon/RiemannToPerm` expects a product of Riemann tensors
# or Riemann and one epsilon. RICCI TENSOR IS NOT ALLOWED.
# Ricci scalar is allowed.
#
# expr must pass through Canonical1 and `Canon/RicciToRiemann commands
#
`Canon/RiemannToPerm` := proc(expr)
options `Copyright 1995-2006 by R Portugal. All rights reserved`;
local i,displ,L,res,L1,L2;

if expr=-1 then
	[-1,[]]

elif expr=R[] then
	[[2,3]]

elif type(expr,R[]^posint) then
	`Canon/prod`([[2,3]],
	seq(`Canon/Canonical/sym_displ`(4*i,[[2,3]]),i=1..op(2,expr)-1))

elif type(expr,'_SCALAR(indexed)') then
	_EnvCanonRiemannToPerm := true;
	`Canon/Canonical/Tensor`(op(expr));

elif type(expr,'_SCALAR(`*`)') then
	_EnvCanonRiemannToPerm := true;
	`Canon/Canonical/TensorProd`(op(expr),_EnvCanonRiemannToInv,_EnvCanonRiemannToPerm);

elif type(expr,'_SCALAR(anything)^posint') then
	L := `Canon/RiemannToPerm`(op(1,expr));
	`Canon/prod`(L,
		seq(`Canon/Canonical/sym_displ`(4*i,L),i=1..op(2,expr)-1))

elif type(expr,`*`) then
	L := [op(expr)];
	if has(L,'epsilon') then
		L1,L2 := selectremove(has,L,'epsilon');
		L := [op(L2),op(L1)]
	end if;
	displ := 0;
	res := NULL;
	for i in L do
		L := `Canon/Canonical/sym_displ`(displ,
					`Canon/RiemannToPerm`(i));
		displ := ceil(max(op(indets(L,integer)))/4)*4;
		res := res,L;
	end do;
	`Canon/prod`(res)

else 
	error "not expecting the expression ",expr

end if;

end proc:



#
# PermToRiemann
#
# ([[2,3,5],[4,7,6]],2,false)  ->  R[a1,a2,a3,a4]*R[-a1,-a2,-a3,-a4]
#
# ([[2,3,5],[4,7,6]],1,true)   ->  R[a1,a2,a3,a4]*epsilon[-a1,-a2,-a3,-a4]
#
# ([[2,3],[6,7]],2,false)      ->  _SCALAR(R[a1,a2,-a1,-a2])*
#                                  _SCALAR(R[a3,a4,-a3,-a4])
#
# dual: false - non dual invariants
#       true -  dual invariants
#
PermToRiemann := proc(g::genperm, dual::boolean)
local degree;

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

degree := ceil(max(op(indets(g,integer)))/4);

if dual then degree := degree-1 end if;

if nargs<>2 then ########3
	error "expecting 2 arguments: g and dual (false or true), got ",nargs

elif dual then
	if max(op(indets(g,integer)))>4*(degree+1) then
		error "expecting a permutation compatible with the degree"
	end if;

elif max(op(indets(g,integer)))>4*degree then
	error "expecting a permutation compatible with the degree"

end if;

`Canon/TensorPrint`(
	`Canon/Canonical/Remove_SCALAR`(
		`Canon/PermToRiemann`(g,degree,dual) ) );

end proc:


`Canon/PermToRiemann` := proc(g,degree,dual)
local L,L_ind,L1,S,S1,expr,deg,sign;

if g[1]=-1 then
	sign := -1
else
	sign := 1
end if;

if dual then
	deg := degree+1;
else 
	deg := degree;
end if;

L := map( `Canon/power`, [$1..4*deg], g );

L_ind := `Canon/MkNewDumInds`(2*deg,{'a0'});

L1 := [seq(L[4*i-3..4*i],i=1..deg)];

S := {seq(L[i]=-(-1)^L[i]*L_ind[ceil(L[i]/2)],i=1..nops(L))};

#
# This splits the invariant in product of subinvariants
#
S1 := subs( S,
	`Canon/tools/joinsets`(
		subs({seq(2*i=2*i-1,i=1..nops(L)/2)}, 
				map( x->{op(x)}, {op(L1)}))
				)
		);

L1 := subs(S,L1);

expr := convert(map(x->R[op(x)],L1[1..degree]), `*`);

if dual then
	expr := expr * epsilon[op(subs(S,L1[-1]))]
end if;

sign*mul('_SCALAR'(select(has,expr,i)), i=S1);

end proc:



#
# InvToPerm
#
# inv: RInv[2,3] or DualRInv[3,4]
#
#
InvToPerm := proc(inv::anyindex(integer,integer))
local z;

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

if op(0,inv) = 'RInv' then
   z := Invars[op(1,inv)][op(2,inv)];
   if not type(z,permutation) then
	z:=max(op(map(op,[indices(Invars[degree])])));
	error "expecting inside invariant bounds", RInv[degree,z], "but got ", 
               inv
   else
	z
   end if;
elif op(0,inv) = 'DualRInv' then
  z := DualInvars[op(1,inv)][op(2,inv)];
  if not type(z,permutation) then
       z:=max(op(map(op,[indices(DualInvars[degree])])));
       error "expecting inside invariant bounds", DualRInv[degree,z],
             "but got ", inv
  else
       z
  end if;

else
  error "expecting type RInv[integer,integer] or DualRInv[integer,integer], got ",
        inv

end if;
end proc:



#
# Convert Weyl to Riemann
#
WeylToRiemann := proc(expr)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

`Canon/TensorPrint`(
	`Canon/WeylToRiemann`( `Canon/TensorUnPrint`(expr) ) )
end proc:


`Canon/WeylToRiemann` := proc(expr)
local S,i,j,k,l,ii,formula,n,expr1;

S := indets(expr,'C[anything,anything,anything,anything]');

n := SpaceDimension;

if n<3 then
	error "expecting SpaceDimension>2, got ",SpaceDimension
end if;

formula := C[i,j,k,l] = R[i,j,k,l] + 1/(n-2)*
	 (g[i,l]*R[j,k]+g[j,k]*R[i,l]-g[i,k]*R[j,l]-g[j,l]*R[i,k]) +
		 1/(n-1)/(n-2)*R[]*(g[i,k]*g[j,l]-g[i,l]*g[j,k]);

expr1 := subs({seq(
	subs({i=op(1,ii),j=op(2,ii),k=op(3,ii),l=op(4,ii)},formula), ii=S)},
	expr );

S := select(hastype,
	indets(expr1,'{_SCALAR(`*`),_SCALAR(`+`),_FREE(`*`),_FREE(`+`)}'),
		'R[anything,anything,anything,anything]');

subs( map( x -> x=Canonical1(op(x),map(eval,CanArg)), S), expr1)

end proc:



#
# Convert Riemann to Weyl
#
RiemannToWeyl := proc(expr)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

`Canon/TensorPrint`(
	`Canon/RiemannToWeyl`( `Canon/TensorUnPrint`(expr) ) )
end proc:


`Canon/RiemannToWeyl` := proc(expr)
local S,i,j,k,l,ii,formula,n,expr1;

S := indets(expr,'R[anything,anything,anything,anything]');

n := SpaceDimension;

if n<3 then
	error "expecting SpaceDimension>2, got ",SpaceDimension
end if;

formula := R[i,j,k,l] = C[i,j,k,l]+1/(n-2)*
		(g[i,k]*R[j,l]-g[i,l]*R[j,k]+g[j,l]*R[i,k]-g[j,k]*R[i,l])+
			1/(n-1)/(n-2)*R[]*(g[i,l]*g[j,k]-g[i,k]*g[j,l]);

expr1 := subs({seq(
	subs({i=op(1,ii),j=op(2,ii),k=op(3,ii),l=op(4,ii)},formula), ii=S)},
	expr );

S := select(hastype,
	indets(expr1,'{_SCALAR(`*`),_SCALAR(`+`),_FREE(`*`),_FREE(`+`)}'),
		'C[anything,anything,anything,anything]');

subs( map( x -> x=Canonical1(op(x),map(eval,CanArg)), S), expr1)

end proc:



#
# Convert Ricci to Riemann
#
RicciToRiemann := proc(expr)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

`Canon/TensorPrint`(
	`Canon/RicciToRiemann`( `Canon/TensorUnPrint`(expr) ) )
end proc:


`Canon/RicciToRiemann` := proc(expr)
local S,inds,i;

S := indets(expr,R[anything,anything]);

inds := `Canon/MkNewDumInds`(nops(S),
		`Canon/abs`(map(op,indets(expr,indexed))));

subs({seq(S[i] = R[op(1,S[i]), inds[i], op(2,S[i]), -inds[i]],i=1..nops(S))},
	expr )

end proc:


#
# Convert Riemann to Ricci
#
RiemannToRicci := proc(expr)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

`Canon/TensorPrint`(
	`Canon/RiemannToRicci`( `Canon/TensorUnPrint`(expr) ) )
end proc:


`Canon/RiemannToRicci` := proc(expr)
local expr1;

# Riemann to Ricci
expr1 := subs( 
      map( x -> if op(1,x)=-op(3,x) then
			x=R[op(2,x),op(4,x)] 
		elif op(2,x)=-op(4,x) then
			x=R[op(1,x),op(3,x)]
		elif op(2,x)=-op(3,x) then
			x=-R[op(1,x),op(4,x)]
		elif op(1,x)=-op(4,x) then
			x=-R[op(2,x),op(3,x)]
		end if,
			indets(expr,R[anything,anything,anything,anything]) ),
      expr );

# Ricci to Ricci Scalar
expr1 := subs(
      map( x -> if op(1,x)=-op(2,x) then 
			x=R[] 
		end if,
			indets(expr1,R[anything,anything]) ), 
      expr1 );

# removing minus sign inside _SCALAR
subs( map( x -> if selectremove(type,op(x),-1)[1]=-1 then
			x=-'_SCALAR'(-op(x))
		end if,
			indets(expr1,'_SCALAR(`*`)')  ),
      '_SCALAR(R[])'=R[],
      expr1 );

end proc:




RiemannToInv := proc(expr)

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

`Canon/TensorPrint`(
	`Canon/Canonical/Remove_SCALAR`(
		`Canon/RiemannToInv`( `Canon/TensorUnPrint`(expr) ) ) )
end proc:



`Canon/RiemannToInv` := proc(expr)
options `Copyright 1995-2006 by R Portugal. All rights reserved`;
local expr0,S;

if type(expr,{set,list,`=`,`+`}) then
	return map(`Canon/RiemannToInv`,expr)
end if;

_EnvCanonRiemannToInv := true;

subs(R[]='RInv[1,1]',
	`Canon/RiemannToRicci`(
		Canonical1(
			`Canon/RicciToRiemann`(expr),map(eval,CanArg) ) ) );

end proc:


#
# InvToRiemann
#
# Input: expr is an tensor expression that may have expression of
#        the kind RInv[degree,n.] or DualRInv[degree,n.].
#
# Output: The RInv[degree,n.] and DualRInv[degree,n.] terms are replaced
#         by the Riemann invariants
#
InvToRiemann := proc(expr) 

options `Copyright 1995-2006 by R Portugal. All rights reserved`;

if type(expr,{set,list,`=`}) then
  return map(procname,expr)
end if;

`Canon/TensorPrint`(
	`Canon/Canonical/Remove_SCALAR`(
		`Canon/InvToRiemann`( 
			`Canon/TensorUnPrint`(expr) ) ) )

end proc:


`Canon/InvToRiemann` := proc(expr)
subs( map( x -> x=`Canon/InvToRiemann/1`(x),
         indets(expr, {'RInv[integer,integer]','DualRInv[integer,integer]'}) ),
      expr)
end proc:


`Canon/InvToRiemann/1` := proc(inv)
local g;
#
# the name of the tables with the permutations are Invars and DualInvars. 
# So Inv + ars = Invars and DualInv + ars = DualInvars
#
g :=  (cat(subs('RInv=Inv','DualRInv=DualInv',op(0,inv)),'`ars`'))[op(1,inv)][op(2,inv)];

if type(g,list) then
	`Canon/PermToRiemann`(g, op(1,inv), evalb(op(0,inv)='DualRInv'))
else 
	inv
end if;

end proc:


#
# `Canon/MkNewDumInds` generates new dummy indices
#
# num is the number of dummy indices generated
# 
# S is a set of non allowed indices
#
# Output: A list containing num dummy indices, 
#         For example, [a0,a1,..]
#
`Canon/MkNewDumInds` := proc(num::nonnegint,S)
local ii,jj,seq1,S1;

global DummyIndices;

if num<1 then return {} end if;

if assigned(DummyIndices) then
   if not type(DummyIndices,'{list(symbol),set(symbol)}') then
	error "expecting DummyIndices a list of symbols, got ",
  	      DummyIndices
   elif nops(DummyIndices)<>nops({op(DummyIndices)}) then
        error "expecting DummyIndices a list of non-repeated symbols, got ",
  	      DummyIndices
   end if;
   S1 := subs(map(x->x=NULL,{op(S)}),DummyIndices);
   if num > nops(S1) then
	seq1 := op(S1);
   else
	return {op(1..num,S1)}
   end if;
else
   seq1 := NULL;
end if;

for ii in '[a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z]' do
    for jj from 0 to 9 do
	if not assigned(``||ii||jj) and not member(``||ii||jj,S) then
		seq1 := seq1,``||ii||jj;
		if nops([seq1])=num then
			return {seq1}
		end if;
	end if;
    end do
end do;

for ii in '[a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z]' do
    for jj from 10 to 99 do
	if not assigned(``||ii||jj) and not member(``||ii||jj,S) then
		seq1 := seq1,``||ii||jj;
		if nops([seq1])=num then
			return {seq1}
		end if;
	end if;
    end do
end do;

for jj from 0 to 10^5 do
    if not assigned(_a||jj) and not member(_a||jj,S) then
		seq1 := seq1,_a||jj;
		if nops([seq1])=num then
			return {seq1}
		end if;
    end if;
end do;

error "warning: could not find a name for a dummy index";

end proc:



#############################################################################
#
# The Canon Package supersedes the old Canon Package
#
#############################################################################
#
# Structure of this file
#
# 1. INITIALIZATION (moved up)
#
# 2. USER PROCEDURES
#
# 3. INTERNAL GROUP PROCEDURES
#
# 4. TENSOR PROCEDURES
#
# 5. TYPES (tensor)
#
# 6. INTERFACE
#
# 7. GENERAL TOOLS
#


#############################################################################
#############################################################################
#
# 2. USER PROCEDURES
#
# Convention: 
#    (1) checks arguments extensively
#    (2) removes the interface
#    (3) calls the respective `Canon/command`
#    (4) puts the interface back
#


#
# TensorDefine
#
#    defines a tensor 
#
# Tensor is an indexed variable: 
#     e.g. R[a,b,-c,-d]
# where R is the tensor name
#       a,b are contravariant indices
#       -c,-d are covariant indices
# 
# Input: N - tensor name (no indices)
#        ni - number of indices
#        KS - list of permutations or a minus-signed permutations that
#             generates the group of symmetries of T
#
# Output: 
#   (1) an equation: TensorTable[tensor,n. of indices][0] = symmetry, base
#   (2) 0 if the symmetry cancels the tensor (e.g. {[[1,2]],[-1,[[1,2]]]})
#
TensorDefine := proc(N::symbol, ni::nonnegint, 
    KS::{tensorsymmetry,symbol},BS::list(posint))

local strongKS,B,i;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if nargs<3 or nargs>4 then
     error "expecting a tensor name (1st arg), number of indices (2nd arg), \
a symmetry (3nd arg), a base (4th arg - optional)"
end if;

if nargs=4 then
     if BS=[] then B:=[$1..ni] else B := BS end if;

elif type(KS,symbol) then
     B := [$1..ni];

else 
     B := NULL  ###sort([op(indets(KS,posint))]);

end if;

forget(`Canon/Canonical/aux`);

if type(KS,symbol) then

     if nargs=4 then
        error "expecting 3 arguments when the 3rd argument is a symbol"
     end if;

     if KS='Symmetric' then

        `Canon/TensorDefine`(N,ni,{seq([[i,i+1]],i=1..ni-1)},B);

     elif KS='AntiSymmetric' then

        `Canon/TensorDefine`(N,ni,{seq([-1,[[i,i+1]]],i=1..ni-1)},B);
     
     else error "expecting a set of permutations, the name Symmetric or AntiSymmetric"

     end if;

else
     if ni<max(op(indets(KS,posint))) then
        error "number of indices not compatible with symmetries"
     end if;
     strongKS := `Canon/schreier-sims`(KS,B);
     `Canon/TensorDefine`(N,ni,strongKS);
end if;
end proc:


#
# CanonicalOne
#
#    returns the canonical form of one tensor or one indexed object
#
# tensor is an indexed variable: 
#     e.g. R[a,b,-c,-d]
# where R is the tensor name
#       a,b are contravariant indices
#       -c,-d are covariant indices
# 
# Input: tensor - a tensor
#        KS - list of permutations or a minus-signed permutations that
#             generates STRONGLY the group of symmetries of tensor
#        Bs - base
#
# Output: tensor with the indices in the canonical ordering
#
`Canon/CanonicalOne` := proc(T::{indexed,_TENSOR(indexed)}, 
          KS::tensorsymmetry, BS::list(nonnegint))

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if nargs<>3 then
     error "expecting a tensor (1st arg), a symmetry (2nd arg), and a base (3rd arg)"
end if;

if KS={} and `Canon/indices`(T)[1]=[] then return T end if;
#
# `Canon/schreier-sims` is not involked
#
`Canon/TensorPrint`(
   `Canon/Canonical/Tensor`(`Canon/TensorUnPrint`(T), KS, BS, 
		_EnvCanonRiemannToInv,_EnvCanonRiemannToPerm ) 
                     )   
end proc:


#
# Canonical
#
#    returns the canonical form of a tensor expression
# 
# Input: expr - a tensor expression
#
# Output: expression with the indices of the tensors in the canonical ordering
#
Canonical := proc(expr::{algebraic,list,set,`=`})

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if nargs<>1 then 
     error "expecting 1 arg, got %1",nargs 
elif not hastype(expr,indexed) then
     return expr
end if;

`Canon/TensorPrint`(
	`Canon/Canonical/Remove_SCALAR`(
		Canonical1(`Canon/TensorUnPrint`(expr),map(eval,CanArg))  )  );

end proc:


#
# Puts the interface
#
TensorPrint := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
`Canon/TensorPrint`(`Canon/TensorUnPrint`(x));
end proc:


#
# Removes the interface
#
TensorUnPrint := proc()
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
`Canon/TensorUnPrint`(args)
end proc:



#############################################################################
#############################################################################
#
# 3. INTERNAL GROUP PROCEDURES

#
# Warning: (1) the internal procedures do not perform type checking.
#          (2) it is expected input with no interface (_TENSOR).
#


#########################################################
# Canon/position                                        #
#                                                       #
# Input: pt - a positive integer possibly in the list   #
#        L - a list                                     #
# Output: the position of the integer in the list or 0  #
#         if not in the list                            #
#########################################################

`Canon/position` := proc(pt,L)

local pos;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if member(pt, L, pos) then pos
                      else 0
end if
end proc:


###########################################################
# Canon/power                                             #
#                                                         #
# Input: po - a position                                  #
#        p - a permutation or a minus-signed permutation  #
# Output: the position of the image                       #
###########################################################

`Canon/power` := proc(po,p)

local pos, c, perm;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if p=[] then RETURN(po) end if;

#
# ignores sign
#

if p[1]=-1 then perm:=p[2]
           else perm:=p
end if;

#
# searches for a cycle containing the element
#

for c in perm while not member(po,c,pos) do end do;

#
# tests for trivial case
#

if not type(pos,posint) then RETURN(po)

#
# tests for end of the cycle
#

elif pos=nops(c) then RETURN(op(1,c))
                 else RETURN(op(pos+1,c))

end if

end proc:


###############################################################
# Canon/prod                                                  #
#                                                             #
# Input: p1, p2 - permutations or a minus-signed permutations #
# Output: product p1xp2, signed only if necessary             #
###############################################################

`Canon/prod` := proc(p1,p2)
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if nargs>2 then
	return `Canon/prod`(`Canon/prod`(p1,p2),args[3..nargs])
end if;

#
# checks trivial case
#

if p1=[] then RETURN(p2)
elif p2=[] then RETURN(p1) 
end if;

#
# output has the appropriate sign
#

if p1[1]=-1 then 
   if p2[1]=-1 then 
      `Canon/prod/nosign`(p1[2],p2[2])
   else 
      [-1,`Canon/prod/nosign`(p1[2],p2)]
   end if;

elif p2[1]=-1 then 
   [-1,`Canon/prod/nosign`(p1,p2[2])]
else 
   `Canon/prod/nosign`(p1,p2)
end if;

end proc:


#########################################
# Canon/prod/nosign                     #
#                                       #
# Input: p1, p2 - unsigned permutations #
# Output: product p1xp2                 #
#########################################

`Canon/prod/nosign` := proc(p1,p2)
local i,elts,res,rescyc,first,image;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# builds a list of nontrivial positions
#

elts := sort([op(indets({op(p1),op(p2)},integer))]);

res := NULL;
rescyc := NULL;

while elts<>[] do

   first:=elts[1];
   rescyc:=elts[1];
   image:=`Canon/power`(`Canon/power`(first,p1),p2);

   #
   # obtains a cycle, stops when it returns to first elt in cycle
   #

   while image<>first do
      rescyc := rescyc,image;
         image := `Canon/power`(`Canon/power`(image,p1),p2);
   end do;

   #
   # removes elements in this cycle from elts
   #

   elts := subs({seq(i=NULL,i=rescyc)}, elts);

   #
   # adds cycle to result
   #

   if nops([rescyc])<>1 then res := res,[rescyc] end if;

end do;

[res];

end proc:


###########################################################
# Canon/g_power                                           #
#                                                         #
# Input: g - a permutation or a minus-signed permutation  #
#        n - an integer                                   #
# Output:  g to power n                                   #
###########################################################

`Canon/g_power` := proc(g,n)
local pos, c, perm;
options remember,system,`Copyright 2006 by R Portugal. All rights reserved`;
if n=0 then []
elif n=1 then g
else `Canon/prod`(procname(g,n-1),g)
end if;
end proc:



#########################################################
# Canon/inv - inverts a permutation                     #
#                                                       #
# Input: p - a permutation or a minus-signed permutation#
# Output: the inverse of p                              #
#########################################################

`Canon/inv` := proc(p)
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if p=[] then 
   []

elif p[1]=-1 then 
   [-1,map(`Canon/invcyc`,p[2])]
else 
   map(`Canon/invcyc`,p)
end if;

end proc:


############################################
# Canon/invcyc - inverts a cycle           #
#                                          #
# Input: c - a cycle                       #
# Output: the inverse cycle                #
############################################

`Canon/invcyc` := proc(c)
local i;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

[c[1], seq( c[nops(c)-i], i=0..nops(c)-2 )];

end proc:



################################################
# Canon/min - minimum with respect to a base #
################################################

`Canon/min` := proc(L,B)

local i;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if L={} then RETURN(NULL) end if;

for i in B do
   if member (i, L) then RETURN (i) end if;
end do;

min(op(L));

end proc:


######################################################################
# Canon/SortPermutations - sorts permutations with respect to a base #
######################################################################

`Canon/SortPermutations` := proc(L,B)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
if nops(L)=1 then 
   L
else 
   _EnvCanonBase := B;
   sort(L,`Canon/SortPermutations/aux`);
end if;

end proc:


#
#  auxiliary procedure        
#

`Canon/SortPermutations/aux` := proc(L1,L2)

local i,p1,p2,n;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if L1 = [] then 'true' 
elif L2 = [] then 'false' 
elif L1=L2 then 'true'
else 

   n := max(op(_EnvCanonBase));

   for i in _EnvCanonBase do

      p1 := `Canon/power`(i,L1);
      if not member(p1,_EnvCanonBase,'p1') then 
         p1 := p1+n 
      end if;

      p2 := `Canon/power`(i,L2);
      if not member(p2,_EnvCanonBase,'p2') then 
         p2 := p2+n 
      end if;

      if p1 < p2 then RETURN('true')
      elif p1 > p2 then RETURN('false')
      end if;

   end do;

   error "base %1 not enough for ordering permutations %2,%3",
              _EnvCanonBase,L1,L2;
end if;

end proc:


##############################################################################
# Canon/orbit - calculates one orbit                                         #
#                                                                            #
# Input: p - position of a point belonging to Omega;                         #
#        K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#                                                                            #
# Output: the orbit positions p^<K>.                                         #
#                                                                            #
##############################################################################

`Canon/orbit` := proc(p,K)

local orb, new_orb, g, orb_pt, Next;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

orb := {p};
readlib('queue');
new_orb:= queue['new'](p);

while not queue['empty'](new_orb) do

   #
   # gets one orbit point
   #

   orb_pt := queue['dequeue'](new_orb);

   #
   # adds the action of all generators on this point to the orbit...
   #

   for g in K do

      Next := `Canon/power`(orb_pt,g);

      #
      # ...provided they are not already in orbit
      #

      if not member(Next,orb) then

         orb := orb union {Next};

         #
         # enqueues the Next points to be analised
         #

         queue['enqueue'](new_orb, Next);

      end if;
   end do;

end do;

orb

end proc:


##############################################################################
# Canon/orbit1 - calculates one orbit, schreier vector                       #
#                       and backward pointers                                #
#                                                                            #
# Input: p - position of any point belonging to Omega;                       #
#        K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#        B - a base for G;                                                   #
#        sch and back - unassigned names for the Schreier vector and         #
#                               backward pointers.                           #
#                                                                            #
# Output: the orbit positions p^<K>.                                         #
#         the Schreier vector and backward pointers components               #
#               for this orbit only                                          #
##############################################################################

`Canon/orbit1` := proc(p,K,B,sch,back)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

`Canon/orbit2`(`Canon/min`(`Canon/orbit`(p,K),B),K,B,sch,back)

end proc:


##############################################################################
# Canon/orbit2 - calculates one orbit, schreier vector and backward          #
#       pointers, starting from the orbit representative                     #
#                                                                            #
# Input: p - an orbit representative belonging to Omega;                     #
#        K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#        B - a base for G;                                                   #
#        sch and back - unassigned names for the Schreier vector and         #
#                               backward pointers.                           #
#                                                                            #
# Output: the orbit positions p^<K>.                                         #
#         the Schreier vector and backward pointers components               #
#               for this orbit only                                          #
##############################################################################

`Canon/orbit2` := proc(p,K,B,sch,back)

local orb, new_orb, g, orb_pt, Next,rep;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

orb := {p};
readlib('queue');
new_orb:= queue['new'](p);

while not queue['empty'](new_orb) do

   #
   # gets one orbit point
   #

   orb_pt := queue['dequeue'](new_orb);

   #
   # adds the action of all generators on this point to the orbit...
   #

   for g in K do

      Next := `Canon/power`(orb_pt,g);

      #
      # ...provided they are not already in orbit
      #

      if not member(Next,orb) then

         orb := orb union {Next};

         #
         # also builds Schreier vector and backward pointers
         #

         sch[Next] := g;
         back[Next] := orb_pt;

         #
         # enqueues the next points to be analised
         #

         queue['enqueue'](new_orb, Next);

      end if;
   end do;
end do;

rep := `Canon/min`(orb,B);
sch[rep]:=0;
back[rep]:=0;

orb

end proc:



##############################################################################
# Canon/allorbits - Calculates all the orbits, schreier vector and backward  #
#                pointers                                                    #
#                                                                            #
# Input: K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#        B - a base for G;                                                   #
#        sch and back - unassigned names for the Schreier vector and         #
#                               backward pointers.                           #
#                                                                            #
# Output: The set of orbits, the Schreier vector and the backward pointers.  #
##############################################################################

`Canon/allorbits` := proc(K, B, sch,back)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local orb, orbits, l, i;

sch := 'sch';
back := 'back';

#
# builds a set of nontrivial positions
#

l:= indets(K,posint);

orbits := {};

while l<>{} do

   orb := `Canon/orbit2`(`Canon/min`(l,B), K, B, sch, back);

   #
   # removes the points on the orbit from the list
   #

   l := subs({ seq(i=NULL, i=op(orb) ) }, l);

   #
   # adds the orbit to the set of orbits
   #

   orbits := orbits union {orb}

end do;

orbits;

end proc:


##############################################################################
# Canon/allorbits0 - Simply calculates all the orbits                        #
#                                                                            #
# Input: K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#                                                                            #
# Output: The set of orbits.                                                 #
##############################################################################

`Canon/allorbits0` := proc(K)

local orb, orbits, l, i;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# builds a set of nontrivial positions
#

l:= indets(K,posint);

orbits := {};

while l<>{} do

   orb := `Canon/orbit`(l[1], K);

   #
   # removes the points on the orbit from the list
   #

   l := subs({ seq(i=NULL, i=op(orb) ) }, l);

   #
   # adds the orbit to the set of orbits
   #

   orbits := orbits union {orb}

end do;

orbits;

end proc:


###############################################################
# Canon/trace - Traces a Schreier vector                      #
#                                                             #
# Input: pos - a position                                     #
#        v - a Schreier vector                                #
#        b - the associated backward pointers                 #
#                                                             #
# Output: an element mapping the orbit representative to pos  #
###############################################################

`Canon/trace` := proc(pos, v, b)

options`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local p, res;

p := pos;
res := [];

#
# Returns identity if v is not known. 
# Useful to trace singleton orbits in `Canon/doubcoset`
#

if type(v[p],indexed) then RETURN([]) end if;

while v[p]<>0 do

   res := `Canon/prod`(v[p],res);
   p := b[p];
end do;

res;

end proc:


###############################################################################
# Canon/singcoset - obtains the canonical representative of a single coset    #
#                                                                             #
# Input: g - any permutation or minus-signed permutation belonging to HxSp+2q #
#               permutations strongly generating a subgroup S of HxSp+2q;     #
#        strongK=[s1,s2,...,sn] - list of permutations or minus-signed        #
#               permutations strongly generating a subgroup S of HxSp+2q;     #
#        B - base for S with respect to strongK                               #
#        free_ind - list of positions of free indices                         #
#                                                                             #
# Output: newK - strong generating set for S that stabillizes the positions   #
#                of the free indices                                          #
#         newB - base for <newK>                                              #
#         canrep - canonical representative of the coset of S in HxSp+2q that #
#                  contains g                                                 #
###############################################################################

`Canon/singcoset` := proc(g,strongK,B,free_ind)

local canrep,newK,delta,i,sch,back,L,p,newB,omega,free,k,N,n;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

canrep:=g;
newK:=strongK;
newB:=B;
free:=free_ind;
N:=nops(free);
n:=0;

for i in B while n<N do

   delta := [op(select(member,
                     `Canon/orbit2`(i, newK, newB, sch, back),free))];
   if delta=[] then next end if;

   L:= map(`Canon/power`, delta , canrep);
   k:=`Canon/min`(L,B);
   p:= delta[`Canon/position`(k,L)];

   omega:=`Canon/trace`(p,sch, back); 
   sch:='sch';
   back:='back';
   canrep:=`Canon/prod`(omega,canrep);

   #
   # the position of free indices changes by the action of omega
   #

   free:=map(`Canon/power`,free,`Canon/inv`(omega));

   newB:=subs(i=NULL,newB);
   newK:=remove(has,newK,i);

   n := n+1;
end do;

return canrep, newK, free;

end proc:


##############################################################################
# Canon/doubcoset - Obtains the canonical representative of a double coset   #
#                                                                            #
# Input: g - any permutation or minus-signed permutation in HxS2n            #
#         strongKS = [s1,s2,...,sn] - list of permutations or minus-signed   #
#               permutations strongly generating a subgroup S of HxS2n;      #
#        baseS - base for S with respect to strongKS                         #
#         strongKD = [d1,d2,...,dn] - list of permutations or a minus-signed #
#               permutations strongly generating the subgroup of symmetries  #
#                of dummy indices                                            #
#         baseD - base for D with respect to strongKD                        #
#         DCSize - an unused name                                            #
#                                                                            #
# Output: doubcanrep - canonical representative of the double coset          #
#                        S x g x D                                           #
##############################################################################

`Canon/doubcoset` := proc(g,strongKS,baseS,strongKD,baseD)

options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local BS,BD,i,KS,KD,T,DeltaD,Deltab,vS,wS,vD,wD,b,alpha,images,p,j,Deltap,
        Next,s,d,L,sinv,gd,L1,F,sgd,verify_zero,DeltaS,points;

if strongKS union strongKD={} then 
        return g
end if;

#
# initializes group S
#

points := indets(strongKD,posint) union {op(baseD)}; 
###7/5/03 empty orbit problem solved

KS := strongKS;
BS := [op(baseS),op(sort([op(remove(member,points,baseS))]))];

KD := strongKD;
BD := baseD;

#
# Checks if g is in S or D
# 

if has(KS,-1) or has(KD,-1) then verify_zero := true 
                            else verify_zero := false
end if;

if not(verify_zero) and nargs<>6 and
(`Canon/isingroup`(g,KS,BS) or `Canon/isingroup`(g,KD,BD)) then

   return []

end if;


#
# initializes table T and alpha
#

T[] := [[],[]];
alpha := {[]};


for i to nops(BS)-1 do

   b := BS[i];

   DeltaD := `Canon/allorbits0`(KD);

   #
   # adds singletons to the orbits
   # 

   DeltaD := DeltaD union map(x->{x},
                remove((x,y)->has(y,x),points,DeltaD));

   DeltaS := `Canon/allorbits`(KS,BS,'vS','wS');

   Deltab := {b};

   for j in DeltaS do

      if member(b,j) then 

         Deltab := j; 
         break;
      end if;
   end do;

   images := map (`Canon/doubcoset/F1`,alpha,T,DeltaD,Deltab,g);

   p[i] := `Canon/min`(images, BS);

   BD:=[p[i], op(subs({p[i]=NULL, p[i]-1=NULL, p[i-1]=NULL}, BD )) ];

   `Canon/allorbits`(KD,BD,'vD','wD');

   for j in DeltaD do

      if member(p[i],j) then 

         Deltap := j; 
         break;
      end if;
   end do;

   for L in alpha do

      L1:=T[op(L)];

      s := L1[1];
      d := L1[2];

      Next := map( `Canon/power`, Deltab, s ) intersect 
                          map( `Canon/power` , Deltap, 
                                  `Canon/inv` (`Canon/prod`(g,d)) ) ;
      sinv := `Canon/inv`(s);
      gd := `Canon/prod`(g,d);

      for j in Next do

          T[op(L),j] := [ `Canon/prod`(
              `Canon/trace`(`Canon/power`(j,sinv),vS,wS),s),
                  `Canon/prod`(d,`Canon/inv`(
                     `Canon/trace`(`Canon/power`(j,gd),vD,wD)))];
      end do;

      T[op(L)] := evaln(T[op(L)]);
   end do;

   alpha := { indices(T) };

   #
   # verifies if the result is zero
   #

   if verify_zero then 

      F := proc(x)

      if x<>[] and x[1]=-1 then x[2] 
                           else x 
      end if
      end;

      sgd := map(`Canon/doubcoset/sgd`,alpha,T,g);

      if nops(sgd)<>nops(map(F,sgd)) then RETURN(0) end if;
   end if;

   KS := remove(has,KS,b); 

#  BS := BS[2..-1];

   KD := remove(has,KD,p[i]);
end do;

#
# if DCSize given, stores the size of the double coset in it
#
#if nargs = 6 then
#   DCSize := group[grouporder](permgroup(nops(BS), strongKS))*
#      group[grouporder](permgroup(nops(BS), strongKD))/nops(alpha);
#end if;

RETURN(`Canon/doubcoset/sgd`(alpha[1],T,g))

end proc:


#
#  F1 - auxiliary procedure
#

`Canon/doubcoset/F1` := proc(L,T,deltaD,deltab,g)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

op( map( op , select(has,deltaD, map( `Canon/power`,deltab,
   `Canon/doubcoset/sgd`(L,T,g)    ) ) ) )
end proc:



#
#  sgd - auxiliary procedure
#

`Canon/doubcoset/sgd` := proc(L,T,g)

local L1;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

L1:=T[op(L)];

`Canon/prod`(`Canon/prod`(L1[1],g),L1[2]);

end proc:


##############################################################################
# Canon/SGS_dummy - Strong Generating Set for symmetry group D (dummy indices)        
#
# Input:
#
# free - positions of free indices
# q - number of pairs of dummy indices
# metric - 1, -1 or 0
#
##############################################################################

`Canon/SGS_dummy` := proc(free,q,metric)

options remember, `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local BD,i;

BD := remove(member,[$1..nops(free)+2*q],free);

if metric=1 then 

   { seq( [[ BD[2*i-1],BD[2*i] ]] , i=1..q ), 
   seq( [[ BD[2*i-3], BD[2*i-1] ],
           [ BD[2*i-2], BD[2*i] ]] , i=2..q ) },BD;

elif metric=-1 then

   { seq( [-1,[[ BD[2*i-1],BD[2*i] ]]] , i=1..q ), 
   seq( [[ BD[2*i-3], BD[2*i-1] ],[ BD[2*i-2], BD[2*i] ]] , i=2..q ) },BD;

elif metric=0 then

   {seq([[BD[2*i-3],BD[2*i-1]],[BD[2*i-2],BD[2*i]]],i=2..q)},BD 

else 
   error "expecting 1,-1, or 0 as 3rd argument (metric), got %1",metric
end if;

end proc:


##############################################################################
# Canon/Canonical/Image - finds canonical configuration in image notation    #
#                                                                            #
# Input: L - a permutation in HxS2n in image notation                        #
#         KS=[s1,s2,...,sn] - list of permutations or minus-signed           #
#               permutations strongly generating a subgroup S of HxS2n;      #
#         p - number of free indices                                         #
#         q - number of PAIRS of dummy indices                               #
#         metric - specifies metric. Can be 1 (symmetric), -1 (antisymmetric)#
#                        or 0 (none)                                         #
#                                                                            #
# Output: the permutation which gives the canonical configuration            #
#                of indices, or 0, in image notation                         #
#                                                                            #
##############################################################################

`Canon/Canonical/Image` := proc(L,KS,BS,p,q,metric)

local g;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

g := `Canon/Canonical/g`(convert(L, 'disjcyc'),KS,BS,p,q,metric);

if g=0 then 
	0
elif g<>[] and g[1]=-1 then
	[-1,convert(g[2],'plist',p+2*q)]
else
	[+1,convert(g,'plist',p+2*q)]
end if

end proc:


##############################################################################
# Canon/Canonical/g - Obtains the canonical configuration permutation        #
#                                                                            #
# Input: g - any permutation or minus-signed permutation in HxS2n            #
#         KS=[s1,s2,...,sn] - list of permutations or minus-signed           #
#               permutations strongly generating a subgroup S of HxS2n;      #
#        BS - base for S with respect to KS                                  #
#         p - number of free indices                                         #
#         q - number of PAIRS of dummy indices                               #
#         metric - specifies metric. Can be 1 (symmetric), -1 (antisymmetric)#
#                        or 0 (none)                                         #
#                                                                            #
# Output: the permutation which gives the canonical configuration            #
#                of indices, or 0                                            #
##############################################################################

`Canon/Canonical/g` := proc(g,KS,BS,p,q,metric)

local KD,BD,res,can_g,h,KS1,free;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

can_g,KS1,free := `Canon/singcoset`(g,KS,BS,
                map(`Canon/power`,[$1..p],`Canon/inv`(g))   );

KD,BD:=`Canon/SGS_dummy` (free,q,metric);

h := convert([op(free) , op(BD)], 'disjcyc');

#
# Here we use BD instead of BS as the base for KS1 because of 2 reasons:
# (1) there is no natural ordering for dummy indices in terms of their names.
# (2) for Riemann monomials, non-standard bases are slower (why?).
#
res := `Canon/doubcoset`( `Canon/prod`(can_g,h), KS1, BD, KD, BD ); 

#
# To use the base given by the user, BD (3rd argument) must be replaced
# by select(member,BS,BD) as below
#
#res := `Canon/doubcoset`( `Canon/prod`(can_g,h), KS1,
#                select(member,BS,BD), KD, BD);  
#

if res=0 then 0
         else `Canon/prod`(res, `Canon/inv`(h))
end if

end proc:


##############################################################################
# Canon/isingroup - Performs membership testing                              #
#                                                                            #
# Input: g - a permutation in Sn                                             #
#         KK - a strong generating set for a subgroup G<Sn                   #
#         B - base for G with respect to KK                                  #
#                                                                            #
# Output: true if g belongs to G, false if not                               #
##############################################################################

`Canon/isingroup`:=proc(g,KK,B)

options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local h,k,K,i,b,o,delta,v,w;

if g=[] or member(g,KK) then 
   RETURN(`true`) 
end if;

h := g;
k := nops(B);
K := KK;

for i to k do

   b := B[i];
   o := `Canon/power`(b,h);
   delta := `Canon/orbit2`(b,K,B,v,w);

   if member(o,delta) then 
      h := `Canon/prod`(h,`Canon/inv`(`Canon/trace`(o,v,w)));
   else 
      RETURN(`false`);
   end if;
        
   if i<>k then 
      K:= remove(has, K, b) 
   end if;
end do;

evalb(h=[]);

end proc:


##############################################################################
# Canon/schreier-sims - Obtains a strong generating set                      #
#                                                                            #
# Input: K - a generating set for a group G                                  #
#          optionaly, a base can be given                                    #
# Output: the base and strong generating set for G                           #
##############################################################################

`Canon/schreier-sims` := proc(K,BB)

options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local B,base,strongK,i,Kbar,l;

if nargs=2 then
   B := BB;
else 
   B := sort([op( map(p-> 
           if p[1] = -1 then p[2][1][1] else p[1][1] end if,K) )]);
end if;

base := op(B);
strongK:=K;

for i from nops(B) to 1 by -1 do

   Kbar := remove(has,K, {op(1..i-1,B)} );
   l:= `Canon/schreier-sims/aux`([base],strongK,i,Kbar); 

   #
   # updates base and K
   #

   base:=base,op(l[2]);
   strongK:=strongK union {op(l[1])};
end do;
 
strongK,[base];

end proc:


#
#  auxiliary procedure        
#

`Canon/schreier-sims/aux`:=proc(base,strongK,i,Kbar)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local B,k,K,newB,newK,K_i,currentgens,oldgens,olddelta,delta,sch,back,
        gamma_,genset,s,gamma_s,trace2,trace1,g,m,j,bj,level,l;

B:=base;
#k:=nops(B);
K:=strongK;
newB:=NULL;
newK:=NULL;

K_i:=remove(has,K,{op(1..i-1,B)});

currentgens:=K_i;
oldgens:=K_i minus Kbar;
olddelta:=`Canon/orbit`(B[i],oldgens);
delta:=`Canon/orbit`(B[i],K_i);

for gamma_ in delta do

   if member(gamma_,olddelta) then genset:=Kbar
                             else genset:=currentgens;
   end if;
       
   for s in genset do

      #
      # uses current schereier vector
      #

      `Canon/allorbits`(currentgens,B,sch,back);
      gamma_s:=`Canon/power`(gamma_,s);
      trace2:=`Canon/trace`(gamma_s,sch,back);
      trace1:=`Canon/trace`(gamma_,sch,back);
      g:=`Canon/prod`(`Canon/prod`(trace1,s),`Canon/inv`(trace2));

      if not(`Canon/isingroup`(g,K,B)) then

         K:=K union {g};
         newK:=newK,g;

         k:=nops(B);

         if has(g,{op(B)}) then 
                                
            for j to k while not has(g,B[j]) do end do;
         else
            j := k+1;

            if g[1]=-1 then if g[2]=[] then
                                bj:=1 ############### check here ###############
                            else bj:= g[2][1][1];
                            end if;
                       else bj:= g[1][1];
            end if;

            B:=[op(B),bj];
            newB:=newB,bj;
         end if;

         for level from j to i+1 by -1 do

            l:=`Canon/schreier-sims/aux`(B,K,level,{g});
                
            #
            # updates B and K
            #

            for m in l[2] do

               if not member(m,B) then
 
                  B:=[op(B),m];
                  newB:=newB,m;
               end if;
            end do;
            newK:=newK,op(l[1]);
            K:=K union {op(l[1])};
         end do;
      end if;
   end do;
end do;

[[newK],[newB]];

end proc:





#############################################################################
#############################################################################
#
# 4. TENSOR PROCEDURES
#
# General Comments:
#
# The output may have the functions _SCALAR and _FREE, which have associated 
# `print/_SCALAR` and `print/_FREE` procedures. The argument of _SCALAR
# must be an indexed object with no free indices or a product of indexed 
# objects forming a scalar (no free indices) and sharing dummy indices
# (a block).
# Examples: _SCALAR(T[a,b,-a,-b]), _SCALAR(V[a]*V[-a]).
#
# The argument of _FREE  must be an indexed object with at least one free index
# with one or more dummy indices. It can be a product with free indices of 
# indexed objects sharing dummy indices.
# Examples: _FREE(T[a,b,d,-d]), _FREE(T[a,b,-c,-d]*T[c,d,-e,-f]*V[e]*V[f]).
#
# In principle the user should not (and need not) to use these inert functions.
# They were in were introduced for the canonicalization procedures and 
# to speed up calculations. To canonicalize each block apart is faster than 
# to canonicalize all tensors together. 
# For example:
# R is the Riemann tensor in this example.
# If the product R[a,b,-a,-b]*R[c,d,-c,-d]*R[e,f,g,h]*R[-e,-f,-g,-h]*
# R[i,j,k,l]*R[-i,-j,-k,-l]*R[m,n,o,p]*R[-m,-o,-n,-p] is converted to
# a permutation, the time to find the canonical representative of the 
# corresponding double coset is huge. This product is equivalent to
# _SCALAR(R[a,b,-a,-b])^2*_SCALAR(R[e,f,g,h]*R[-e,-f,-g,-h])^2*
# _SCALAR(R[m,n,o,p]*R[-m,-o,-n,-p]), which is canonicalized in less than
# a second, because each block is canonicalized separately. 
#
# In the procedure TensorProd,
# the blocks wrapped by either the inert function _SCALAR or _FREE
# merge into single tensors of the form:
#    CanonTensor[T1,n1,T2,n2,...][seq of all indices]
# where T1, T2, ... are the tensor names and n1, n2, ... are
# the number of indices - they are sorted inside TensorProd.
#
# In the procedure CanDummy, the index names are replaced by canonical names.
# This is performed at the algebraic level inside procedure Canonical
# so it simplifies expressions of the following type:
#    A[a,b]*B[-a,-b]*A[c,d]*B[-c,-d] -> 
#         CanonTensor[A,2,B,2][_1_3,_2_4,-_1_3,-_2_4]^2
#


#
# Defines a tensor - internal procedure
#
`Canon/TensorDefine` := proc(N,ni,sym,base)
global TensorTable;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if member([-1,[]],sym) then 
     0
else 
     TensorTable[N,ni][0] := sym,base;
     return evaln(TensorTable[N,ni][0])=sym,base;
end if;

end proc:


#
# Canonicalizes an expression - Algebraic level.
# 
# In order to remember properly, call Canonical1(...,map(eval,CanArg))
#
Canonical1 := proc(expr1)

local expr;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# Types list,set,`=` are allowed. Types `+`,`*`,`^` are expected.
#
if type(expr1,'{list,set,`=`}') then
     return map(Canonical1,expr1,map(eval,CanArg))
end if;

expr := subs(Canonical2(expr1));

subs( map(
	x->x=`Canon/Canonical/CanonTensorToProd`(x),
		indets(expr,'CanonTensor')),
			expr);

end proc:


#
# Canonical2 retuns a sequence of 2 operands:  S, expr_out
# S is the set of equations to replace canonical names for dummies
# expr_out is an expression with CanonTensor and dummies with 
# canonical names.
# 
Canonical2 := proc(expr1)

local i,expr,S,S1,S2,F,n1,n2;
global CanonCanDummyInds, CanonExtraDummyInds;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

expr := expr1;

#
# Checks indices as a side effect
#
EnvCanonFreeInds := {op(`Canon/indices`(expr)[2])}; 

#
# Finds all dummy indices.
#
_EnvCanonDummyInds := `Canon/AllDummies`(expr);

#
# Returns here if there are no dummies
#
if nops(_EnvCanonDummyInds)=0 then 
   return {},subs(map(x->x=`Canon/Canonical/Tensor`(x),indets(expr,indexed)),expr) 
end if;

#
# unwrapping _FREE(anything) and expanding
#
CanonExtraDummyInds := {};

expr := subs(
	  map(x->x=`Canon/Canonical/NewDummies`(op(x)),
		indets(expr,'_FREE(`*`)')),
			expr);

expr := subs(map(x->x=op(x), indets(expr,'_FREE(indexed)')), expr);

expr := `Canon/Canonical/expand`(expr);

#
# Stores all dummy indices with canonical names. Example: _1_2
#
CanonCanDummyInds := {};

if assigned(CanonSimp) then
	if not type(CanonSimp,procedure) then
		lprint("*** warning: expecting CanonSimp to be a function");
	end if;
	expr := CanonSimp(`Canon/Canonical/aux`(expr));
else
	expr := `Canon/Canonical/aux`(expr);
end if;

#
# Replaces canonical names for dummies by the names used in the input
#
n1 := nops(CanonCanDummyInds);
n2 := nops(_EnvCanonDummyInds);
if n1>n2 then
	_EnvCanonDummyInds := _EnvCanonDummyInds union
		`Canon/MkNewDumInds`(n1-n2, _EnvCanonDummyInds union
			map(`Canon/abs`,map(op,indets(expr1,indexed) ) )
					);
end if;

S := {seq(CanonCanDummyInds[i]=_EnvCanonDummyInds[i],i=1..n1)};

unassign('CanonCanDummyInds,CanonExtraDummyInds');

return S,expr

end proc:




#
# `Canon/Canonical/aux` - Procedure auxiliar to Canonical2 (recursive procedure).
# 
`Canon/Canonical/aux` := proc(expr)

local expr1;
global CanonExtraDummyInds;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if type(expr,indexed) then
     if has(expr,_EnvCanonDummyInds union CanonExtraDummyInds) then
		`Canon/Canonical/CanDummy`(expr)
     else
		`Canon/Canonical/Tensor`(expr)
     end if

elif type(expr,'{_SCALAR(anything),_FREE(anything)}') then 

     `Canon/Canonical/CanDummy`(op(expr))

elif type(expr,function) then

     expr #########

elif type(expr,`*`) then  

     expr1 := `Canon/Canonical/SplitBlocks`(expr,
				_EnvCanonDummyInds union CanonExtraDummyInds,
				EnvCanonFreeInds  );

     if type(expr1,`*`) then
	      map(`Canon/Canonical/aux`,expr1)
     else
	      `Canon/Canonical/aux`(expr1)
     end if;

elif type(expr,{`+`,`^`}) then

     map(`Canon/Canonical/aux`, expr)

else expr

end if;

end proc:


#
# Splits the product in blocks of terms that share dummy indices.
# No canonicalization is performed.
# Wrap blocks sharing only dummy indices with the inert function _SCALAR.
# Wrap blocks with free indices (and at least a pair of dummy indices)
# with the inert function _FREE.
# This is very important for performance.
# E.g.
#      A[a]*B[-a]*A[b]*B[-b] ->  _SCALAR(A[a]*B[-a])*_SCALAR(A[b]*B[-b])
#      T[i,a,-a]    -> _FREE(T[i,a,-a])  (no dummy renaming)
#      V[i,a]*W[-a] -> _FREE(V[i,a]*W[-a])
#
`Canon/Canonical/SplitBlocks` := proc(expr)
options `Copyright 2003 by R Portugal. All rights reserved`;
local i,j,L,F,S,S1,expr1,expr2;

expr1,expr2 := selectremove(x->type(x,indexed) and 
		has(x,_EnvCanonDummyInds union CanonExtraDummyInds), expr);

if not type(expr1,`*`) then 
	return expr 
end if;

#
# NewDummies modifies CanonExtraDummyInds
#
L := map(`Canon/Canonical/NewDummies`,[op(expr1)]);

F := proc(x,j,i) if has(x,j) then i else NULL end if end proc:

S := subs({}=NULL,`Canon/tools/joinsets`(
		{seq({seq(F(L[i],j,i),i=1..nops(L))},
			j=_EnvCanonDummyInds union CanonExtraDummyInds)}));

S,S1 := selectremove(x->nops(x)>1,S);

S1 := map(op,S1) union ({$1..nops(L)} minus map(op,S));

F := proc(x) 
       if has(x,EnvCanonFreeInds) then 
       	     '_FREE'(x) 
       else '_SCALAR(x)'
       end if 
     end proc;

mul(L[i],i=S1)*mul(F(mul(L[j],j=i)),i=S)*expr2;

end proc:


#
# Replaces dummy indices by canonical names after putting indices in
# the canonical position. The canonical names are of the form: 
# _PositonOfi_PositionOf-i 
#       (e.g. CanonTensor[V,2][i,-i]  ->  CanonTensor[V,2][_1_2,-_1_2])
# This replacement is used at the algebraic level, so the
# following expression is simplified:
#       V[i,-i]*V[j,-j]  ->  V[_1_2,-_1_2]^2
#
# Examples 
#   (1)   Input:  V[i,-i]
#         Output: _SCALAR(V[_1_2,-_1_2])
#
#   (2)   Input:  V[k,i,-i]
#         Output: _FREE(V[k,_2_3,-_2_3])
#
#   (3)   Input:  A[a]*B[-a]  (or [A[a],B[-a]])
#         Output: CanonTensor[A,1,B,1][_1_2,-_1_2]
#
`Canon/Canonical/CanDummy` := proc(expr)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local L,i,S,T,sign;
global CanonCanDummyInds;

if type(expr,indexed) then
     T := `Canon/Canonical/Tensor`(expr)
elif type(expr,`*`) then
     T := `Canon/Canonical/TensorProd`(expr,_EnvCanonRiemannToInv,_EnvCanonRiemannToPerm)
end if;

if T=0 then
   return 0
end if;

if type(-T,indexed) then
   sign := -1;
   T := -T;
else sign := 1;
end if;

L := remove(member,[op(T)],EnvCanonFreeInds);

S := {seq(i=evaln(_||(`Canon/position`(i,L))||
           ('_')||(`Canon/position`(-i,L))), 
                 i=`Canon/indices`(T)[1])};

CanonCanDummyInds := CanonCanDummyInds union map(rhs,S);

if type(T,'{CanonTensor,RInv[integer,integer],DualRInv[integer,integer]}') then
	sign*subs(S,T)
elif has(T,EnvCanonFreeInds) then
	sign*'_FREE'(subs(S,T))
else
	sign*'_SCALAR'(subs(S,T))
end if;

end proc:


#
# `Canon/Canonical/TensorProd` 
#            canonicalize a block of tensors (tensor product)
#
# INPUT: L is a list of at least 2 tensors sharing dummy indices. 
# OUTPUT: 0 or T[inds] or -T[inds], the tensor indices in the canonical 
#         ordering.
#
# Output for RInv and DualRInv (Invariants): RInv[2,3] ...
#
`Canon/Canonical/TensorProd` := proc(Tensors1,EnvCanonRiemannToInv)

local T,Tensors,N_Tensors,TensorNames,L_N_I,NewNames,SameNames,s_subsop1,
s_subsop2,i,pos,N,p,sym_T,MergedTensors,N_MergedTensors,L_all_ind,L_N_I1,
F,L_displ,merged_sym,merged_base,Merged_T,TensorSign,S,S_inv,j,jj;

global TensorTable;

options remember,system,`Copyright 1997-2003 by R Portugal All rights reserved`;

#
# sorts Tensors according to tensor names and number of indices
#
Tensors := sort([op(Tensors1)],
			proc(a,b)
				local L;
                        L := map(x->evaln(_||(op(0,x))||('_')||(nops(x))), [a,b]);
                        evalb(L=sort(L))
			end proc);

N_Tensors := nops(Tensors);
TensorNames  := map2(op,0,Tensors);
L_N_I        := map(nops,Tensors); # List of Number of Indices 
NewNames   := [seq(evaln(Canon_||(TensorNames[i])||('_')||
                      (L_N_I[i])), i=1..N_Tensors)];
SameNames := `Canon/selectrepeated`(NewNames);

#
# Merging equal tensors (same name and same n. of indices)
#
s_subsop1   := NULL;
s_subsop2   := NULL;

for i in SameNames do
  pos := [`Canon/positions`(i,NewNames)];
  N   := nops(pos);
  p   := L_N_I[pos[1]];

  if not `Canon/has_sym`(Tensors[pos[1]],'sym_T') then 
       sym_T := {},[$1..p];
  end if;

  `Canon/TensorDefine`(i,p*N,`Canon/Canonical/MkSym`(p,N,sym_T));

  s_subsop1 := s_subsop1, seq(j=NULL,j=pos[2..-1]),
           pos[1]=i[seq(op(Tensors[jj]),jj=pos)];
  s_subsop2 := s_subsop2, seq(j=NULL,j=pos[2..-1]),
           pos[1]=i;
end do;

MergedTensors  := subsop(s_subsop1, Tensors);
N_MergedTensors := nops(MergedTensors);
NewNames := subsop(s_subsop2, NewNames);

#
# Full reduction: all tensors merge to form one tensor: Merged_T
#
L_all_ind := map(op,MergedTensors);
L_N_I1 := map(nops,MergedTensors);

#
# calculates the displacement
#
F := proc(L)
 options remember,system;
  if L=[] then 
       0
  else
       F(L[1..-2]),op(-1,[F(L[1..-2])])+L[-1]
  end if
end proc:

L_displ := [F(L_N_I1[1..-2])];

F := proc(x) if type(x,set) then x else {} end if; end proc:

merged_sym := {seq(op(`Canon/Canonical/sym_displ`(L_displ[i],
                F(TensorTable[op(0,MergedTensors[i]),L_N_I1[i]][0][1]))),
                   i=1..N_MergedTensors)};

F := proc(x) if type(x,list) then x else [] end if; end proc:

merged_base := [seq(op(`Canon/Canonical/sym_displ`(L_displ[i],
                F(TensorTable[op(0,MergedTensors[i]),L_N_I1[i]][0][2]))),
                   i=1..N_MergedTensors)];


#
# Clear the symmetries of the tensors defined above
#
for i to N_MergedTensors do
   if not member(op(0,MergedTensors[i]),TensorNames) then
   #
   # The next command is the proper command to unassing. It is not
   # working. This seems to be a bug in unassign.
   #
   # unassign(evaln(TensorTable[op(0,MergedTensors[i]),L_N_I1[i]]));
   TensorTable := table(remove(has,op(eval(TensorTable)),
                                    op(0,MergedTensors[i])));
   end if
end do;


# Converts to CanonTensor blocks of 2 or more tensors.
#      A[a]*B[-a] ->  CanonTensor[A,1,B,1][a,-a]
#
T := CanonTensor[op(map(x->(op(0,x),nops(x)),Tensors))];

#
# RInv and DualRInv
#
if EnvCanonRiemannToInv=true then

	if TensorNames=['R'$N_Tensors] and {op(L_N_I)}={4} and
			`Canon/FreeIndices`(T[op(L_all_ind)])={} then
		_EnvCanonDual:=false;
	elif TensorNames=['R'$(N_Tensors-1),'epsilon'] and {op(L_N_I)}={4} and
			`Canon/FreeIndices`(T[op(L_all_ind)])={} then
		_EnvCanonDual:=true;	
	end if;	

end if;

`Canon/Canonical/Tensor`(T[op(L_all_ind)],merged_sym,merged_base,
	_EnvCanonRiemannToInv,_EnvCanonRiemannToPerm);

end proc:




#
# canonicalizes single tensors
#
`Canon/Canonical/Tensor` := proc(T,KS,BS,EnvCanonRiemannToInv,EnvCanonRiemannToPerm)

options remember,system,`Copyright 2003 by R Portugal. All rights reserved`;
local  symm,i,L,ind,p,q,can_L,free,dummy;

if nargs=1 then
	symm := TensorTable[op(0,T),nops(T)][0];
	if nops([symm])=2 then   
		return `Canon/Canonical/Tensor`(T,symm,
			_EnvCanonRiemannToInv,_EnvCanonRiemannToPerm) # 5 nargs
	else
		return `Canon/Canonical/Tensor`(T,{},[$1..nops(T)],
			_EnvCanonRiemannToInv,_EnvCanonRiemannToPerm)
	end if;
end if;

# 
# free and p store information on the free indices 
# dummy and q store information on the dummy indices
#
ind := `Canon/indices`(T);
p := nops(ind[2]);
q := nops(ind[1]);

if q=0 and KS={} then return(T) end if;

free := `Canon/tools/sort`(ind[2]);
dummy := sort(ind[1]);

#
# L stores the positions of indices
#
L := subs({seq(free[i]=i,i=1..p)},
        {seq(-dummy[i]=p+2*i,i=1..q)},
                {seq(dummy[i]=p+2*i-1,i=1..q)},[op(T)]);
#
# calls computational group algorithms
#
can_L := `Canon/Canonical/Image`(L,KS,BS,p,q,1);

if can_L=0 then RETURN(0) end if;

#
# RInv and DualRInv
#
if EnvCanonRiemannToInv=true and assigned(_EnvCanonDual) then 

	L := `Canon/PermToInv`(convert(can_L[2],'disjcyc'),_EnvCanonDual);

	if not type(L,list) then 
		return can_L[1]*L
	end if;

elif EnvCanonRiemannToPerm=true then

	if can_L[1] = -1 then
		return [-1,convert(can_L[2],'disjcyc')]
	else
		return convert(can_L[2],'disjcyc')
	end if;

end if;

#
# restores user-level indices and outputs the result
#
L := subs({seq(i=free[i],i=1..p)},
        {seq(p+2*i=-dummy[i],i=1..q)},
                {seq(p+2*i-1=dummy[i],i=1..q)},can_L[2]);

can_L[1]*op(0,T)[op(L)];

end proc:


#
# Converts 
#        CanonTensor[T1,n1,T2,n2,...][seq of indices]
# back to 
#        T1[...]*T2[...]*...
#
# Example:
#          Input:  CanonTensor[A,1,B,1][_1_2,-_1_2]
#          Output: A[a]*B[-a]
#
`Canon/Canonical/CanonTensorToProd` := proc(x)
options remember,system,`Copyright 2003 by R Portugal. All rights reserved`;
if has(x,EnvCanonFreeInds) then
	'_FREE'(`Canon/Canonical/CanonTensorToProd/1`([op(op(0,x))],[op(x)]))
else
	'_SCALAR'(`Canon/Canonical/CanonTensorToProd/1`([op(op(0,x))],[op(x)]))
end if;
end proc:

`Canon/Canonical/CanonTensorToProd/1` := proc(L,ind)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
if L=[] then
     1
else 
     L[1][op(1..L[2],ind)]*procname(L[3..-1],ind[L[2]+1..-1])
end if;
end proc:


#
# Returns true if a tensor has symmetry and false otherwise.
# If a 2nd argument is given, the symmetry is assigned to it.
#
`Canon/has_sym` := proc(expr,p)
local n,symmetry;
global TensorTable;
options `Copyright 1997 by R Portugal All rights reserved`;
n:=nops(expr);
if n<2 
  then false
  else symmetry := TensorTable[op(0,expr),n][0];
       if nops([symmetry])=2 then
            if nargs=2 then p := symmetry end if;
            true
       else 
            false
       end if
end if
end proc:


#
# `Canon/Canonical/MkSym`
#
#      creates symmetry for product of equal tensors
#
# Input: n - number of indices of each tensor
#        N - number of equal tensors
#        sym_T - symmetry of the original tensor
# 
# Output: a strong generating set of the symmetry of the product of tensors 
#
`Canon/Canonical/MkSym` := proc(n,N,sym,Base)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local i,j,base;

if Base=[] then 
     base := [$1..n]
else
     base := Base
end if;

{ seq(op(`Canon/Canonical/sym_displ`(j*n,sym)),j=0..N-1),
     seq([seq([i+j*n,n+i+j*n],i=1..n)],j=0..N-2)  },
[ seq(op(`Canon/Canonical/sym_displ`(j*n,base)),j=0..N-1)  ]

end proc:


#
# `Canon/Canonical/sym_displ`
#
#      add i to the symmetry
#
# Input: i - non negative integer
#        sym_T - symmetry
# 
# Output: add i to all permutations 
#
`Canon/Canonical/sym_displ` := proc(i,sym_T)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

subs(map(proc(x) x=x+i end proc,indets(sym_T,posint)),sym_T)

end proc:


#
# `Canon/Canonical/NewDummies` creates local dummy index names
# for expression inside _FREE(...)
#
`Canon/Canonical/NewDummies` := proc(expr)
local S;
global CanonExtraDummyInds;
options `Copyright 2006 by R Portugal. All rights reserved`;
S := {seq(i=readlib('`tools/gensym`')(i),i=`Canon/AllDummies`(expr))};
CanonExtraDummyInds := CanonExtraDummyInds union map(rhs,S);
subs(S,expr)
end proc:



#
# `Canon/indices`
#
#      returns the indices of a tensor or tensor expression
#
# Input: expr
# Output: [ [dummy indices], [free indices], [numerical indices] ]
#
# Attention: The output does not contain all indices necessarily, dummy and 
# numerical indices can be missing in expr of type "+". Eg. T[i,-i,1]+T[j,-j,2]
#
`Canon/indices` := proc(expr) 

local nameind,repeated,ind,numind,s; 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;

if nargs=0 or not hastype(expr,indexed) then RETURN([[],[],[]]) end if;

ind:=[`Canon/indices/rec`(expr)];

nameind:=select(type,ind,'nameindex');
numind :=select(type,ind,'numericindex');

#
# Next line does not allow the presence of numerical indices
# in the Canon package.
#
if nops(numind)>0 then 
  error "expecting no numerical indices, but got %1",numind 
end if;

if nops(numind)+nops(nameind)<>nops(ind) then 
  error "there are indices of wrong type: ",op(ind)
end if;

if nops({op(nameind)})<>nops(nameind) then 
  ERROR(`wrong choice of indices`) 
end if;

repeated := select(type,subs(0=NULL,subs(
              map(proc(x) x=0 end,{op(convert(nameind,`+`))}),
                nameind)),'symbol');

s := subs(0=NULL,subs( map(proc(x) x=0 end,repeated) ,nameind));

[repeated, s, numind]

end proc:


`Canon/indices/rec` := proc(expr)  
local i,ind,ind1,op0,n,s;
options remember,system,`Copyright 1997-2003 by R Portugal. All rights reserved`;
if type(expr,'indexed') and not member(op(0,expr),'{RInv,DualRInv}')
   then op(expr)
elif type(expr,'function')
   then op0:=op(0,expr);
        if member(op0,'{_SCALAR,_FREE}')
          then procname(op(expr))
        elif member(op0,'{diff,Diff}')
          then ind:=`Canon/indices`(op(1,expr));
               ind1:=procname(op(2,expr));
               if ind1=NULL 
                 then  op(op(1,ind)),op(map(proc(x) -x end,op(1,ind))),
                             op(op(2,ind)),op(op(3,ind))
               elif member(op0,'{diff,Diff}') and
                    (nops([ind1])>1 or type(ind1,'covariantindex') or 
                     (type(-ind1,'symbol') and has(ind,-ind1)))
                 then ERROR(`wrong choice of indices in derivative`) 
               else  op(op(1,ind)),op(map(proc(x) -x end,op(1,ind))),
                             op(op(2,ind)),op(op(3,ind)),-ind1
               end if
        elif type(op0,'function') 
          then s:=op(0,op0);
               if s='D' then ind:=`Canon/indices`(op(op0)); 
                             op(op(2,ind)),op(op(3,ind))
               elif type(s,'indexed') and op(0,s)='D'
                 then if hastype(s,'covariantindex')
                        then ERROR(`wrong choice of indices in D operator`)
                      end if;
                      ind:=`Canon/indices`(op(op0));
                      op(op(2,ind)),op(op(3,ind)),op(map(x->-x,[op(ind)]))
               elif not hastype(s,'indexed')
                 then NULL
               else ERROR(`NOT IMPLEMENTED`)
               end if 
        elif nops(expr)=1
           then ind:=seq(op(i),i=op(2..3,`Canon/indices`(op(expr))));
                if ind=NULL then NULL
                else userinfo(1,'Canon',`assuming that function `||op0||
                      ` is linear in its argument`);
                     ind
                end if
        elif nops(expr) > 1
          then if hastype(expr,'constanttensor')
               then ind:=seq(op(i),i=op(2..3,`Canon/indices`(op(1,expr))));
                    if ind=NULL then NULL
                    else userinfo(1,'Canon',`assuming that function `||op0||
                                    ` has tensors in its 1st argument only`);
                         ind
                    end if
               else NULL
               end if
        else NULL         
        end if
elif type(expr,`*`) 
   then ind1:=seq(`Canon/indices`(i),i=expr);
        ind:=seq(op(op(2,i)),i=ind1);
        if {seq(op(op(1,i)),i=ind1)} intersect {seq(`Canon/abs`(i),i=ind)} <> {}
          then ERROR(`dummy index equal to free index`) 
        end if;
        ind,seq(op(op(3,i)),i=ind1)
elif type(expr,`+`)
   then ind:=map(proc(x) op(2..3,`Canon/indices`(x)) end,[op(expr)]);
        n:=nops(ind)/2;
        if not(nops({seq({op(op(2*i-1,ind))},i=1..n)})=1   )
        then `Canon/TensorPrint`(expr);
             ERROR(`free indices do not match or summing tensors of different type`) 
        end if;
        procname(op(1,expr))
elif type(expr,`^`) 
   then ind:=map(`Canon/indices`,[op(expr)]);
        if op(2,op(1,ind))=[] and op(2,op(2,ind))=[]
          then NULL
        else  `Canon/TensorPrint`(expr);
              ERROR(`free indices in power`)
        end if
else NULL
end if 
end proc:


#
# finds all dummy indices of an expression
#
`Canon/AllDummies` := proc(expr)
options `Copyright 2006 by R Portugal. All rights reserved`;
select(type,map(op,indets(expr,indexed)),symbol) minus 
	`Canon/FreeIndices`(expr)
end proc:


#
# `Canon/FreeIndices` finds the absolute values of the free indices
#
# This procedure is faster than `Canon/Indices`. No mathematical
# checking is performed.
#
`Canon/FreeIndices` := proc(expr)
local T,i,expr1;
options remember,system,`Copyright 2006 by R. Portugal. All right reserved.`;
if type(expr,indexed) then
	select(type,`Canon/abs`(subs(0=NULL,{op(`+`(op(expr)))})),symbol)

elif type(expr,`+`) then
	`Canon/FreeIndices`(op(1,expr))

elif type(expr,`*`) then
	expr1 := `Canon/FreeIndices/aux`(expr);
	if expr1=1 then
		{}
	elif type(expr1,indexed) then
		`Canon/FreeIndices`(expr1)
	elif type(expr1,`*`) then
		`Canon/FreeIndices`(T[seq(op(i),i=expr1)])
        else
		error "expecting type `*`, got ",expr1
	end if;

elif type(expr,'_FREE(anything)') then
	`Canon/FreeIndices`(op(expr))

elif type(expr,function) then
	map(`Canon/abs`,{op(`Canon/indices`(expr)[2])})

else 
	{}

end if;
end proc:


`Canon/FreeIndices/aux` := proc(expr)
local T;
options remember,system,`Copyright 2006 by R. Portugal. All right reserved.`;
if type(expr,indexed) and nops(expr)>0 then
	expr

elif type(expr,`+`) then
	`Canon/FreeIndices/aux`(op(1,expr))

elif type(expr,`*`) then
	map(`Canon/FreeIndices/aux`,expr)

elif type(expr,'_SCALAR(anything)') then
	1

elif type(expr,'_FREE(anything)') then
	`Canon/FreeIndices/aux`(op(expr))

elif type(expr,function) then
	T[op(`Canon/indices`(expr)[2])]

else 
	1

end if;
end proc:


#
# `Canon/Canonical/expand` expands only terms with indices
#
`Canon/Canonical/expand` := proc(expr)
local e0,e1,e2,dumind,i,S;
options remember,system,`Copyright 2006 by R. Portugal. All right reserved.`;
if type(expr,{`+`,`^`}) then
   map(`Canon/Canonical/expand`,expr)

elif type(expr,`*`) then
   e0 := map(`Canon/Canonical/expand`,expr);
   S := {};
   for i to nops(e0) do
       e1 := op(i,e0);
       if type(e1,`+`)  then
           e2 := subsop(i=1,e0);
           S := `Canon/FreeIndices`(e1) intersect `Canon/FreeIndices`(e2);
           if S <> {} then
                break;
           end if;
        end if;
   end do;
   if S <> {} then
       if type(e2,`*`) then
            e2,e0 := selectremove(has,e2,S);
       else e0 := 1;
       end if;
       procname(e0*map(`*`,e1,e2))
   else
       e0
   end if

else 
   expr

end if;
end proc:


#
# `Canon/Canonical/Perm`
#
# Input: g - sigperm or permutation
#        degree - number of Riemann tensors
# 
# Output: canonical permutation
#
`Canon/Canonical/Perm` := proc(g,degree,dual)
local KS,BS,CanonMetricSymmetry;
options `Copyright 2006 R Portugal. All rights reserved`;

CanonMetricSymmetry := 1; # allows future extensions - antisymmetric metric

if dual then

	KS,BS := `Canon/Canonical/MkSymRieProd*epsilon`(degree);
	`Canon/Canonical/g`(g,KS,BS,0,2*(degree+1),CanonMetricSymmetry);

else
	KS,BS := `Canon/Canonical/MkSymRieProd`(degree);
	`Canon/Canonical/g`(g,KS,BS,0,2*degree,CanonMetricSymmetry);

end if;
end proc:


#
# `Canon/Canonical/MkSymRieProd`
#
#      creates symmetry for product of Riemann tensors
#
# Input: degree - number of Riemann tensors
#        sym_T - symmetry of the original tensor
# 
# Output: a strong generating set of the symmetry of the product of 
# Riemann tensors 
#
`Canon/Canonical/MkSymRieProd` := proc(degree)

options `Copyright 2006 R Portugal. All rights reserved`;

`Canon/Canonical/MkSym`(4,degree,TensorTable[R,4][0],[$1..4*degree])

end proc:


#
# `Canon/Canonical/MkSymRieProd*epsilon`
#
#      creates symmetry for product of Riemann tensors and an epsilon
#
# Input: degree - number of Riemann tensors
#        sym_T - symmetry of the original tensor
# 
# Output: a strong generating set of the symmetry of the product of 
# Riemann tensors and an epsilon
#
`Canon/Canonical/MkSymRieProd*epsilon` := proc(degree)
local KS,BS;
options `Copyright 2006 R Portugal. All rights reserved`;

KS,BS := `Canon/Canonical/MkSymRieProd`(degree);

KS union `Canon/Canonical/sym_displ`(4*degree,TensorTable[epsilon,4][0][1]), 
[op(BS),op(`Canon/Canonical/sym_displ`(4*degree,TensorTable[epsilon,4][0][2]))]

end proc:


#############################################################################
#############################################################################
#

#############################################################################
#############################################################################
#
# 6. INTERFACE



`Canon/TensorPrint` := proc(expr)
local op0;
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if type(expr,'indexed') then 
     if member(op(0,expr),'{RInv,DualRInv}') then
           expr
     else
           _TENSOR(expr)
     end if;
elif type(expr,'function') then 
     op0:=op(0,expr);
     if op0='_SCALAR' then
          '_SCALAR'(procname(op(expr)))
     elif op0='diff' or op0='Diff' then 
          op0(op(map(procname,[op(expr)])))
     elif type(op0,'indexed') then 
          _TENSOR(expr)
     elif has(op0,'D') then 
          op(0,op0)(procname(op(op0)))(op(map(procname,[op(expr)])))
     else op0(op(map(procname,[op(expr)])))
     end if
elif type(expr,{`+`,`*`,`=`,list,set}) then 
     map(procname,expr)
elif type(expr,`^`) then 
     if type(op(1,expr),'constanttensor') and nops(op(1,expr))<>0 and 
                   op(2,expr)<>-1 then 
                   procname(op(1,expr))^procname(op(2,expr))
     else map(procname,expr)
     end if
else expr
end if
end proc:


`_TENSOR/F` := proc(x)
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
#
# Maple 9 requires ` ` instead of `` for representing upper indices 
#
   if x=1 then readlib('`tools/gensym`')(` `)^`1`
   elif x=-1 then readlib('`tools/gensym`')(``)[`1`]
   elif type(x,'contravariantindex') then readlib('`tools/gensym`')(` `)^x
   elif type(x,'covariantindex') then readlib('`tools/gensym`')(``)[-x]
   else error "%1 is not an index of correct type",x
   end if
end proc:


`Canon/TensorUnPrint` := proc(expr)
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
subs(map(x->x=op(x),indets(expr,'_TENSOR(anything)')),expr);
end proc:


`Canon/Canonical/Remove_SCALAR` := proc(expr)
options `Copyright 2006 by R Portugal. All rights reserved`;

if type(expr,'{_SCALAR(anything),_FREE(anything)}') then

	op(expr)

elif type(expr,`+`) then

	map(procname,expr)

else 
	expr

#
# This part was removed since expression such as 
# (R[a,b]*R[-a,-b]+1)*(R[a,b]*R[-a,-b]+2)
# generates an error when expanded.
#
#else
#	subs( map( x -> x=
#		map( y ->  if type(y,'{_SCALAR(anything),_FREE(anything)}') then
#				op(y)
#			   else
#				y
#			   end if, x ),
#				indets(expr,'specop(anything,`+`)') ) , 
#		expr )

end if

end proc:


#############################################################################
#############################################################################
#
# 7. GENERAL TOOLS


#
# Canon/positions - positions of an element occuring in a list
#
`Canon/positions` := proc(i,L)

options `Copyright 1997-2003 by R Portugal. All rights reserved`;
local p,j;

if member(i,L,'p')
        then p,seq(p+j,j=procname(i,L[p+1..-1]))
        else NULL
end if

end proc:


#
# `Canon/selectrepeated` 
#
#         selects the repeated elements of a list preserving the order
#
# Example:  [A,A,A,B,C,C,D,E] -> [A,C]
#
`Canon/selectrepeated` := proc(L)
local f1,L1;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
f1 := proc(x,L) local pos; member(x,L,pos); pos=NULL end;
L1 := {op(subsop(op(map(f1,{op(L)},L)),L))};
map(op,sort([op(map(`Canon/position`,L1,L))]),L);
end proc:


#
# Sorts L using lexicographical order for contravariant indices followed
# by the covariant indices:
# [ a, b, c, ...,  -d, -e, -f, ...] 
#
# Numeric indices is not allowed.
#
`Canon/tools/sort` := proc(L) 
local L_contra,L_co;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;

L_contra,L_co := selectremove(type,L,'contravariantnameindex');

[op(sort(L_contra,'lexorder')),
   op(map(x->-x,sort(map(x->-x,L_co),'lexorder')))]

end proc:


#
# input: a list, a set or an index
#
`Canon/abs` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if type(x,{'list,set'}) then map(procname,x)
elif type(x,{'symbol','posint'}) then x
elif type(-x,{'symbol','posint'}) then -x
elif x=0
  then error "zero cannot be used as index"
else error "wrong index type %1",x 
end if
end proc:


#
# Join sets that have at least one element in common.
# E.g.: {{6},{1,2},{2,3},{4,5},{6,7},{4,7,8}}  ->
#                      {{4, 5, 6, 7, 8}, {1, 2, 3}}
#
`Canon/tools/joinsets` := proc(S)
local S1,S2,s;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;

if nops(S)<2
  then S
else S1 := procname(S[2..-1]);
     s  := S[1];
     S2  := select(has,S1,s);
     {s union map(op,S2)} union remove(member,S1,S2)
end if;
end proc:


setup()
end module:


protect( 'Invar' ):

##### END OF CANON PACKAGE #####




if not assigned(InvarDirectory) then
	InvarDirectory := kernelopts('homedir'):
	lprint(cat("InvarDirectory not assigned, assuming ",InvarDirectory));
	lprint("Try:  InvarDirectory := c:/directory_name; ");
end if:
#
# Non duals
#
read cat(InvarDirectory,"/Invar1.mpl"):
read cat(InvarDirectory,"/Invar2.mpl"):
read cat(InvarDirectory,"/Invar3.mpl"):
read cat(InvarDirectory,"/Invar4.mpl"):
#
# Duals
#
read cat(InvarDirectory,"/Invar1Dual.mpl"):
read cat(InvarDirectory,"/Invar2Dual.mpl"):
read cat(InvarDirectory,"/Invar3Dual.mpl"):
#
# Reduction of Nonduals to Duals 
#
read cat(InvarDirectory,"/Invar4.mpl"):

#infolevel[Canon] := 5;



